"""This module is for computing CCA and CKA."""

from enum import Enum

import pydra
import torch


def compute_cca(features_x: torch.Tensor, features_y: torch.Tensor) -> float:
    """Compute the mean squared CCA correlation (R^2_{CCA}).

    Args:
        features_x (torch.Tensor): A num_examples x num_features matrix of features.
        features_y (torch.Tensor): A num_examples x num_features matrix of features.

    Returns:
        float: The mean squared CCA correlations between X and Y.
    """
    qx, _ = torch.linalg.qr(features_x)
    qy, _ = torch.linalg.qr(features_y)
    result = torch.norm(qx.t() @ qy) ** 2 / min(features_x.shape[1], features_y.shape[1])
    return result.item() if isinstance(result, torch.Tensor) else float(result)


class CKAKernelType(Enum):
    """CKA kernel types."""

    LINEAR = "linear"
    RBF = "rbf"


def compute_cka(
    features_x: torch.Tensor,
    features_y: torch.Tensor,
    kernel: CKAKernelType = CKAKernelType.LINEAR,
    threshold: float = 1.0,
) -> float:
    """Compute CKA between feature matrices.

    Args:
        features_x (torch.Tensor): A num_examples x num_features matrix of features.
        features_y (torch.Tensor): A num_examples x num_features matrix of features.
        kernel (CKAKernelType): Type of kernel to use (CKAKernelType.LINEAR or CKAKernelType.RBF).
        Default is CKAKernelType.LINEAR.
        threshold (float): Fraction of median Euclidean distance to use as RBF kernel bandwidth
            (used only if kernel is CKAKernelType.RBF).

    Returns:
        float: The value of CKA between X and Y.
    """

    def _gram_linear(x: torch.Tensor) -> torch.Tensor:
        """Compute Gram (kernel) matrix for a linear kernel.

        Args:
            x (torch.Tensor): A num_examples x num_features matrix of features.

        Returns:
            torch.Tensor: A num_examples x num_examples Gram matrix of examples.
        """
        return x @ x.t()

    def _gram_rbf(x: torch.Tensor, threshold: float = 1.0) -> torch.Tensor:
        """Compute Gram (kernel) matrix for an RBF kernel.

        Args:
            x (torch.Tensor): A num_examples x num_features matrix of features.
            threshold (float): Fraction of median Euclidean distance to use as RBF kernel bandwidth.

        Returns:
            torch.Tensor: A num_examples x num_examples Gram matrix of examples.
        """
        dot_products = x @ x.t()
        sq_norms = torch.diag(dot_products)
        sq_distances = -2 * dot_products + sq_norms[:, None] + sq_norms[None, :]
        sq_median_distance = torch.median(sq_distances)
        return torch.exp(-sq_distances / (2 * threshold**2 * sq_median_distance))

    def _center_gram(gram: torch.Tensor) -> torch.Tensor:
        """Center a symmetric Gram matrix.

        This is equivalent to centering the (possibly infinite-dimensional) features
        induced by the kernel before computing the Gram matrix.

        Args:
            gram (torch.Tensor): A num_examples x num_examples symmetric matrix.

        Returns:
            torch.Tensor: A symmetric matrix with centered columns and rows.

        Raises:
            ValueError: If the input is not a symmetric matrix.
        """
        if not torch.allclose(gram, gram.t()):
            raise ValueError("Input must be a symmetric matrix.")

        n = gram.size(0)
        unit = torch.ones(n, n, device=gram.device)
        eye = torch.eye(n, device=gram.device)
        unit = unit / n
        haitch = eye - unit
        centered_gram = haitch.mm(gram).mm(haitch)
        return centered_gram

    def _cka(gram_x: torch.Tensor, gram_y: torch.Tensor) -> torch.Tensor:
        """Compute CKA.

        Args:
            gram_x (torch.Tensor): A num_examples x num_examples Gram matrix.
            gram_y (torch.Tensor): A num_examples x num_examples Gram matrix.

        Returns:
            float: The value of CKA between X and Y.
        """
        gram_x = _center_gram(gram_x)
        gram_y = _center_gram(gram_y)

        scaled_hsic = torch.sum(gram_x * gram_y)

        normalization_x = torch.norm(gram_x)
        normalization_y = torch.norm(gram_y)
        return scaled_hsic / (normalization_x * normalization_y)

    if kernel == CKAKernelType.LINEAR:
        gram_x = _gram_linear(features_x)
        gram_y = _gram_linear(features_y)
    elif kernel == CKAKernelType.RBF:
        gram_x = _gram_rbf(features_x, threshold)
        gram_y = _gram_rbf(features_y, threshold)
    else:
        raise ValueError("Unsupported kernel type. Use CKAKernelType.LINEAR or CKAKernelType.RBF.")

    result = _cka(gram_x, gram_y)
    return result.item() if isinstance(result, torch.Tensor) else float(result)


compute_cca_pt = pydra.mark.task(compute_cca)
compute_cka_pt = pydra.mark.task(compute_cka)



"""This module implements some utilities for the voice cloning task."""

from typing import Any, List, Optional

from senselab.audio.data_structures.audio import Audio
from senselab.audio.tasks.voice_cloning.knnvc import KNNVC
from senselab.utils.data_structures.device import DeviceType
from senselab.utils.data_structures.model import SenselabModel, TorchModel


def clone_voices(
    source_audios: List[Audio],
    target_audios: List[Audio],
    model: SenselabModel = TorchModel(path_or_uri="bshall/knn-vc", revision="master"),
    device: Optional[DeviceType] = None,
    **kwargs: Any,  # noqa:ANN401
) -> List[Audio]:
    """Clones voices from source audios to target audios using the given model.

    This function performs pairwise voice cloning, where the voice from each audio sample
    in the `source_audios` list is transformed into the corresponding audio
    sample in the `target_audios` list. The resulting list contains audio samples that
    preserve the content of the original source audio but with the voice replaced by the
    voice from the corresponding target audio.

    Args:
        source_audios (List[Audio]): A list of audio samples whose voices will be "replaced"
            by the voices from the corresponding target audio samples. The content
            (e.g., words) will remain the same, but the voice sounds like the target.
        target_audios (List[Audio]): A list of audio samples whose voices will be extracted
            and used to replace the voices in the corresponding source audio samples.
        model (SenselabModel, optional): The model to use for voice cloning. Currently,
            only KNNVC (K-Nearest Neighbors Voice Conversion) is supported, encapsulated
            by the `TorchModel` class. `TorchModel` is a child class of `SenselabModel`
            and specifies the model and revision for cloning. Defaults to
            `TorchModel(path_or_uri="bshall/knn-vc", revision="master")`.
        device (Optional[DeviceType], optional): The device to run the model on (e.g., CPU or GPU).
            Defaults to None.
        **kwargs: Additional keyword arguments for model-specific parameters that will
            be passed directly to the underlying model's voice cloning method.

    Returns:
        List[Audio]: A list of audio samples with cloned voices from the corresponding source and target audios.

    Raises:
        ValueError: If the lengths of `source_audios` and `target_audios` do not match.
        NotImplementedError: If the specified model is not supported. Currently, only KNNVC is supported.

    Examples:
        >>> source_audios = [Audio.from_filepath("source1.wav"), Audio.from_filepath("source2.wav")]
        >>> target_audios = [Audio.from_filepath("target1.wav"), Audio.from_filepath("target2.wav")]
        >>> cloned_audios = clone_voices(source_audios, target_audios)

    Todo:
        Add logging with timestamps.
    """
    if len(source_audios) != len(target_audios):
        raise ValueError("The list of source and target audios must have the same length.")

    if isinstance(model, TorchModel) and model.path_or_uri == "bshall/knn-vc":
        return KNNVC.clone_voices_with_knn_vc(
            source_audios=source_audios, target_audios=target_audios, model=model, device=device, **kwargs
        )
    else:
        raise NotImplementedError("Only KNNVC is supported for now.")



"""Transcribes audio files with timestamps."""

from typing import List

import pydra

from senselab.audio.data_structures.audio import Audio
from senselab.audio.tasks.forced_alignment.constants import SAMPLE_RATE
from senselab.audio.tasks.forced_alignment.forced_alignment import (
    align_transcriptions,
)
from senselab.audio.tasks.preprocessing.preprocessing import downmix_audios_to_mono, resample_audios
from senselab.audio.tasks.speech_to_text.api import transcribe_audios
from senselab.utils.data_structures.language import Language
from senselab.utils.data_structures.model import HFModel
from senselab.utils.data_structures.script_line import ScriptLine
from senselab.utils.tasks.batching import batch_list


def transcribe_timestamped(
    audios: List[Audio],
    model: HFModel = HFModel(path_or_uri="openai/whisper-tiny"),
    language: Language = Language(language_code="en"),
    n_batches: int = 1,
) -> List[List[ScriptLine]]:
    """Transcribes a list of audio files and timestamps them using forced alignment.

    This function processes the given list of Audio objects by performing
    necessary preprocessing steps (such as downmixing channels and resampling),
    transcribes the audio using the specified speech-to-text model, and applies
    forced alignment to generate a list of ScriptLine objects with timestamps.

    Args:
        audios (list[Audio]): List of Audio objects to be transcribed and
                              timestamped.
        model (HFModel, optional): A Huggingface model for speech-to-text.
                                   Defaults to 'whisper'.
        language (Language, optional): Language object for the transcription.
                                       If None, language detection is triggered
                                       for the 'whisper' model. Defaults to None.
        n_batches (int, optional): The number of batches to split over in the
                                   workflow.

    Returns:
        List[List[ScriptLine]]: List of ScriptLine objects resulting from the
                          transcriptions with timestamps.
    """
    if not audios:
        raise ValueError("The list of audios is empty.")

    for i in range(len(audios)):
        if audios[i].waveform.shape[0] > 1:
            audios[i] = downmix_audios_to_mono(audios=[audios[i]])[0]
        if audios[i].sampling_rate != SAMPLE_RATE:
            audios[i] = resample_audios(audios=[audios[i]], resample_rate=SAMPLE_RATE)[0]

    batched_audios = batch_list(items=audios, n_batches=n_batches)

    wf = pydra.Workflow(
        name="wf",
        input_spec=["batched_audios", "model", "language"],
        batched_audios=batched_audios,
        model=model,
        language=language,
        cache_dir=None,
    )

    @pydra.mark.task
    def transcribe_task(audios: List[Audio], model: HFModel, language: Language) -> List[tuple]:
        transcriptions = transcribe_audios(audios=audios, model=model, language=language)
        return list(zip(audios, transcriptions, [language] * len(audios)))

    wf.add(
        transcribe_task(
            name="transcribe",
            audios=wf.lzin.batched_audios,
            model=wf.lzin.model,
            language=wf.lzin.language,
        )
    ).split("batched_audios", batched_audios=wf.inputs.batched_audios)

    align_transcriptions_task = pydra.mark.task(align_transcriptions)
    wf.add(
        align_transcriptions_task(
            name="align",
            audios_and_transcriptions_and_language=wf.transcribe.lzout.out,
        )
    )

    wf.set_output(
        [
            ("transcriptions", wf.transcribe.lzout.out),
            ("aligned_transcriptions", wf.align.lzout.out),
        ]
    )

    # Execute the workflow
    with pydra.Submitter(plugin="cf") as sub:
        sub(wf)

    return wf.result()[0].output.aligned_transcriptions



"""This module contains functions for plotting audio-related data."""

import matplotlib.pyplot as plt
import torch
from IPython.display import Audio, display

from senselab.audio.data_structures.audio import Audio as AudioData


def plot_waveform(audio: AudioData, title: str = "Waveform", fast: bool = False) -> None:
    """Plots the waveform of an Audio object.

    Args:
        audio (AudioData): An instance of Audio containing waveform data and sampling rate.
        title (str): Title of the plot.
        fast (bool): If True, plots a downsampled version for a faster but less detailed view.

    Todo:
        - Add option to save the plot
        - Add option to choose the size of the Figure
    """
    waveform = audio.waveform
    sample_rate = audio.sampling_rate

    if fast:
        # Downsampling the waveform and sample rate for a quicker plot
        waveform = waveform[::10]

    num_channels, num_frames = waveform.shape
    time_axis = torch.linspace(0, num_frames / sample_rate, num_frames)

    figure, axes = plt.subplots(num_channels, 1, figsize=(12, num_channels * 2))
    if num_channels == 1:
        axes = [axes]  # Ensure axes is iterable
    for c, ax in enumerate(axes):
        ax.plot(time_axis, waveform[c].numpy(), linewidth=1)
        ax.set_ylabel(f"Channel {c + 1}")
        ax.grid(True)

    figure.suptitle(title)
    plt.xlabel("Time [s]")
    plt.show(block=False)


def plot_specgram(spectrogram: torch.Tensor, sample_rate: int, title: str = "Spectrogram") -> None:
    """Plots the spectrogram of an Audio object.

    Args:
        spectrogram (torch.Tensor): A tensor representing the spectrogram.
        sample_rate (int): The sampling rate of the audio data.
        title (str): Title of the spectrogram plot.

    Todo:
        - Add option to save the plot
        - Add option to choose the size of the Figure
    """
    if spectrogram.dim() != 2:
        raise ValueError("Spectrogram must be a 2D tensor.")
    plt.figure(figsize=(10, 4))
    plt.imshow(
        spectrogram.numpy(),
        aspect="auto",
        origin="lower",
        extent=(0, spectrogram.size(1) / sample_rate, 0, sample_rate / 2),
        cmap="viridis",
    )
    plt.colorbar(label="Magnitude (dB)")
    plt.title(title)
    plt.ylabel("Frequency [Hz]")
    plt.xlabel("Time [Sec]")
    plt.show(block=False)


def play_audio(audio: AudioData) -> None:
    """Plays an audio file.

    Args:
        audio (AudioData): An instance of Audio containing waveform data and sampling rate.

    Raises:
        ValueError: If the number of channels is more than 2.
    """
    waveform = audio.waveform.numpy()
    sample_rate = audio.sampling_rate

    num_channels = waveform.shape[0]
    if num_channels == 1:
        display(Audio(waveform[0], rate=sample_rate))
    elif num_channels == 2:
        display(Audio((waveform[0], waveform[1]), rate=sample_rate))
    else:
        raise ValueError("Waveform with more than 2 channels are not supported.")



"""Provides data structures for align functionality."""

from dataclasses import dataclass
from typing import List, Optional, Tuple, TypedDict


class SingleWordSegment(TypedDict):
    """A single word of a speech."""

    word: str
    start: Optional[float]
    end: Optional[float]
    score: float


class SingleCharSegment(TypedDict):
    """A single char of a speech."""

    char: str
    start: Optional[float]
    end: Optional[float]
    score: float


class SingleSegment(TypedDict, total=False):
    """A single segment (up to multiple sentences) of a speech."""

    start: float
    end: float
    text: str
    clean_char: Optional[List[str]]
    clean_cdx: Optional[List[int]]
    clean_wdx: Optional[List[int]]
    sentence_spans: Optional[List[Tuple[int, int]]]


class SingleAlignedSegment(TypedDict):
    """A single segment (up to multiple sentences) of a speech with word alignment."""

    start: float
    end: float
    text: str
    words: List[SingleWordSegment]
    chars: Optional[List[SingleCharSegment]]


class TranscriptionResult(TypedDict):
    """A list of segments and word segments of a speech."""

    segments: List[SingleSegment]
    language: str


class AlignedTranscriptionResult(TypedDict):
    """A list of segments and word segments of a speech."""

    segments: List[SingleAlignedSegment]
    word_segments: List[SingleWordSegment]


@dataclass
class Point:
    """Represents a point in the alignment path.

    Attributes:
        token_index (int): The index of the token in the transcript.
        time_index (int): The index of the time frame in the audio.
        score (float): The alignment score for this point.
    """

    token_index: int
    time_index: int
    score: float


@dataclass
class Segment:
    """Represents a segment of aligned text.

    Attributes:
        label (str): The text label of the segment.
        start (int): The start time index of the segment.
        end (int): The end time index of the segment.
        score (float): The alignment score for the segment.
    """

    label: str
    start: int
    end: int
    score: float

    def __repr__(self) -> str:
        """Provides a string representation of the segment.

        Returns:
            str: A string representation of the segment.
        """
        return f"{self.label}\t({self.score:4.2f}): [{self.start:5d}, {self.end:5d})"

    @property
    def length(self) -> int:
        """Calculates the length of the segment."""
        return self.end - self.start



"""This module provides the implementation of torchaudio utilities for audio features extraction."""

from typing import Dict, List, Optional

import pydra
import torch
import torchaudio

from senselab.audio.data_structures.audio import Audio


def extract_spectrogram_from_audios(
    audios: List[Audio],
    n_fft: int = 400,
    win_length: Optional[int] = None,
    hop_length: Optional[int] = None,
) -> List[Dict[str, torch.Tensor]]:
    """Extract spectrograms from a list of audio objects.

    Args:
        audios (List[Audio]): List of Audio objects.
        n_fft (int): Size of FFT, creates n_fft // 2 + 1 bins. Default is 400.
        win_length (int): Window size. Default is None, using n_fft.
        hop_length (int): Length of hop between STFT windows. Default is None, using win_length // 2.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing spectrograms.
    """
    if win_length is None:
        win_length = n_fft
    if hop_length is None:
        hop_length = win_length // 2
    spectrogram = torchaudio.transforms.Spectrogram(
        n_fft=n_fft,
        win_length=win_length,
        hop_length=hop_length,
    )
    spectrograms = []
    for audio in audios:
        spectrograms.append({"spectrogram": spectrogram(audio.waveform).squeeze(0)})
    return spectrograms


def extract_mel_spectrogram_from_audios(
    audios: List[Audio],
    n_fft: Optional[int] = 400,
    win_length: Optional[int] = None,
    hop_length: Optional[int] = None,
    n_mels: int = 128,
) -> List[Dict[str, torch.Tensor]]:
    """Extract mel spectrograms from a list of audio objects.

    Args:
        audios (List[Audio]): List of Audio objects.
        n_fft (int): Size of FFT, creates n_fft // 2 + 1 bins. Default is 400.
        win_length (int): Window size. Default is None, using n_fft.
        hop_length (int): Length of hop between STFT windows. Default is None, using win_length // 2.
        n_mels (int): Number of mel filter banks. Default is 128.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing mel spectrograms.
    """
    if win_length is None:
        win_length = n_fft
    if hop_length is None:
        if win_length is not None:
            hop_length = win_length // 2
        else:
            raise ValueError("win_length cannot be None")
    mel_spectrograms = []
    for audio in audios:
        mel_spectrogram = torchaudio.transforms.MelSpectrogram(
            sample_rate=audio.sampling_rate,
            n_fft=n_fft,
            win_length=win_length,
            hop_length=hop_length,
            n_mels=n_mels,
        )(audio.waveform)
        mel_spectrograms.append({"mel_spectrogram": mel_spectrogram.squeeze(0)})
    return mel_spectrograms


def extract_mfcc_from_audios(
    audios: List[Audio],
    n_mfcc: int = 40,
    n_ftt: Optional[int] = 400,
    win_length: Optional[int] = None,
    hop_length: Optional[int] = None,
    n_mels: int = 128,
) -> List[Dict[str, torch.Tensor]]:
    """Extract MFCCs from a list of audio objects.

    Args:
        audios (List[Audio]): List of Audio objects.
        n_mfcc (int): Number of MFCCs to return. Default is 40.
        n_ftt (int): Size of FFT, creates n_ftt // 2 + 1 bins. Default is 400.
        win_length (int): Window size. Default is None, using n_ftt.
        hop_length (int): Length of hop between STFT windows. Default is None, using win_length // 2.
        n_mels (int): Number of mel filter banks. Default is 128.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing MFCCs.
    """
    if win_length is None:
        win_length = n_ftt
    if hop_length is None:
        if win_length is not None:
            hop_length = win_length // 2
        else:
            raise ValueError("win_length cannot be None")
    mfccs = []
    for audio in audios:
        mfcc_transform = torchaudio.transforms.MFCC(
            sample_rate=audio.sampling_rate,
            n_mfcc=n_mfcc,
            melkwargs={"n_fft": n_ftt, "win_length": win_length, "hop_length": hop_length, "n_mels": n_mels},
        )
        mfccs.append({"mfcc": mfcc_transform(audio.waveform).squeeze(0)})
    return mfccs


def extract_mel_filter_bank_from_audios(
    audios: List[Audio],
    n_mels: int = 128,
    n_stft: int = 201,
    n_fft: int = 400,
    win_length: Optional[int] = None,
    hop_length: Optional[int] = None,
) -> List[Dict[str, torch.Tensor]]:
    """Extract mel filter bank from a list of audio objects.

    Args:
        audios (List[Audio]): List of Audio objects.
        n_mels (int): Number of mel filter banks. Default is 128.
        n_stft (int): Number of bins in STFT. Default is 201.
        n_fft (int): Size of FFT, creates n_fft // 2 + 1 bins. Default is 400.
        win_length (int): Window size. Default is None, using n_fft.
        hop_length (int): Length of hop between STFT windows. Default is None, using win_length // 2.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing mel filter banks.
    """
    if win_length is None:
        win_length = n_fft
    if hop_length is None:
        hop_length = win_length // 2

    spectrograms = extract_spectrogram_from_audios(audios, n_fft, win_length, hop_length)

    mel_filter_banks = []
    for i, audio in enumerate(audios):
        melscale_transform = torchaudio.transforms.MelScale(
            sample_rate=audio.sampling_rate, n_mels=n_mels, n_stft=n_stft
        )
        mel_filter_banks.append({"mel_filter_bank": melscale_transform(spectrograms[i]["spectrogram"]).squeeze(0)})
    return mel_filter_banks


def extract_pitch_from_audios(
    audios: List[Audio], freq_low: int = 85, freq_high: int = 3400
) -> List[Dict[str, torch.Tensor]]:
    """Extract pitch from a list of audio objects.

    Pitch is detected using the detect_pitch_frequency function from torchaudio.
    It is implemented using normalized cross-correlation function and median smoothing.

    Args:
        audios (List[Audio]): List of Audio objects.
        freq_low (int): Lowest frequency that can be detected (Hz). Default is 85.
        freq_high (int): Highest frequency that can be detected (Hz). Default is 3400.

    Returns:
        List[Dict[str, torch.Tensor]]: List of Dict objects containing pitches.
    """
    pitches = []
    for audio in audios:
        pitches.append(
            {
                "pitch": torchaudio.functional.detect_pitch_frequency(
                    audio.waveform, sample_rate=audio.sampling_rate, freq_low=freq_low, freq_high=freq_high
                ).squeeze(0)
            }
        )
    return pitches


extract_spectrogram_from_audios_pt = pydra.mark.task(extract_spectrogram_from_audios)
extract_mel_spectrogram_from_audios_pt = pydra.mark.task(extract_mel_spectrogram_from_audios)
extract_mfcc_from_audios_pt = pydra.mark.task(extract_mfcc_from_audios)
extract_mel_filter_bank_from_audios_pt = pydra.mark.task(extract_mel_filter_bank_from_audios)
extract_pitch_from_audios_pt = pydra.mark.task(extract_pitch_from_audios)



"""This module implements some utilities for computing the Equal Error Rate (EER)."""

from typing import Tuple

import pydra
import torch
from speechbrain.utils.metric_stats import EER


def compute_eer(predictions: torch.Tensor, targets: torch.Tensor) -> Tuple[float, float]:
    """Compute the Equal Error Rate (EER).

    Args:
        predictions (torch.Tensor): A 1D tensor of predictions.
        targets (torch.Tensor): A 1D tensor of targets.

    Returns:
        Tuple[float, float]: The EER and the threshold for the EER.
    """
    return EER(predictions, targets)


compute_eer_pt = pydra.mark.task(compute_eer)



"""Provides constants for align functionality."""

from senselab.utils.data_structures.model import HFModel, TorchAudioModel

SAMPLE_RATE = 16000

MINIMUM_SEGMENT_SIZE = 400

PUNKT_ABBREVIATIONS = ["dr", "vs", "mr", "mrs", "prof"]

LANGUAGES_WITHOUT_SPACES = ["ja", "zh"]

DEFAULT_ALIGN_MODELS_TORCH = {
    "fr": TorchAudioModel(path_or_uri="VOXPOPULI_ASR_BASE_10K_FR", revision="main"),
    "de": TorchAudioModel(path_or_uri="VOXPOPULI_ASR_BASE_10K_DE", revision="main"),
    "es": TorchAudioModel(path_or_uri="VOXPOPULI_ASR_BASE_10K_ES", revision="main"),
    "it": TorchAudioModel(path_or_uri="VOXPOPULI_ASR_BASE_10K_IT", revision="main"),
}

DEFAULT_ALIGN_MODELS_HF = {
    "en": HFModel(path_or_uri="facebook/wav2vec2-base-960h", revision="main"),
    "ja": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-japanese", revision="main"),
    "zh": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-chinese-zh-cn", revision="main"),
    "nl": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-dutch", revision="main"),
    "uk": HFModel(path_or_uri="Yehor/wav2vec2-xls-r-300m-uk-with-small-lm", revision="main"),
    "pt": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-portuguese", revision="main"),
    "ar": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-arabic", revision="main"),
    "cs": HFModel(path_or_uri="comodoro/wav2vec2-xls-r-300m-cs-250", revision="main"),
    "ru": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-russian", revision="main"),
    "pl": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-polish", revision="main"),
    "hu": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-hungarian", revision="main"),
    "fi": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-finnish", revision="main"),
    "fa": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-persian", revision="main"),
    "el": HFModel(path_or_uri="jonatasgrosman/wav2vec2-large-xlsr-53-greek", revision="main"),
    "tr": HFModel(path_or_uri="mpoyraz/wav2vec2-xls-r-300m-cv7-turkish", revision="main"),
    "da": HFModel(path_or_uri="saattrupdan/wav2vec2-xls-r-300m-ftspeech", revision="main"),
    "he": HFModel(path_or_uri="imvladikon/wav2vec2-xls-r-300m-hebrew", revision="main"),
    "vi": HFModel(path_or_uri="nguyenvulebinh/wav2vec2-base-vi", revision="main"),
    "ko": HFModel(path_or_uri="kresnik/wav2vec2-large-xlsr-korean", revision="main"),
    "ur": HFModel(path_or_uri="kingabzpro/wav2vec2-large-xls-r-300m-Urdu", revision="main"),
    "te": HFModel(path_or_uri="anuragshas/wav2vec2-large-xlsr-53-telugu", revision="main"),
    "hi": HFModel(path_or_uri="theainerd/Wav2Vec2-large-xlsr-hindi", revision="main"),
    "ca": HFModel(path_or_uri="softcatala/wav2vec2-large-xlsr-catala", revision="main"),
    "ml": HFModel(path_or_uri="gvs/wav2vec2-large-xlsr-malayalam", revision="main"),
    "no": HFModel(path_or_uri="NbAiLab/nb-wav2vec2-1b-bokmaal", revision="main"),
    "nn": HFModel(path_or_uri="NbAiLab/nb-wav2vec2-300m-nynorsk", revision="main"),
}



"""This module implements some utilities for audio data augmentation."""

from typing import List, Union

import pydra
import torch
from torch_audiomentations import Compose

from senselab.audio.data_structures.audio import (
    Audio,
    batch_audios,
    unbatch_audios,
)
from senselab.utils.data_structures.device import DeviceType, _select_device_and_dtype


def augment_audios(
    audios: List[Audio], augmentation: Compose, device_options: Union[DeviceType, List[DeviceType]] = [DeviceType.CPU]
) -> List[Audio]:
    """Augments all provided audios with a given augmentation, either individually or all batched together.

    Augment all audios with a user defined augmentation that can be a composition of multiple augmentations. This
    augmentation is either performed on each audio individually or all of the audios provided are batched together
    and run at once. NOTE: if batching, all audios must have the same sampling rate.

    Args:
        audios: List of Audios whose data will be augmented with the given augmentations
        augmentation: A Composition of augmentations to run on each audio (uses torch-audiomentations), should have its
            output_type set to "dict"
        device_options: The device, or a List of possible devices, to use for augmenting. If the chosen device
            is MPS or CUDA then the audios are all batched together, so for optimal performance, batching should
            be done by passing a batch_size worth of audios ar a time

    Returns:
        List of audios that has passed the all of input audios through the provided augmentation. This does
            not necessarily mean that the augmentation has been run on every audio. For more information,
            see the torch-audiomentations documentation.
    """
    augmentation.output_type = "dict"
    new_audios = []
    device_type, dtype = _select_device_and_dtype(
        compatible_devices=device_options if isinstance(device_options, List) else [device_options]
    )
    if device_type == DeviceType.CPU:
        for audio in audios:
            audio_to_augment = audio.waveform.unsqueeze(0)
            augmented_audio = augmentation(audio_to_augment, sample_rate=audio.sampling_rate).samples
            new_audios.append(
                Audio(
                    waveform=torch.squeeze(augmented_audio),
                    sampling_rate=audio.sampling_rate,
                    metadata=audio.metadata.copy(),
                    orig_path_or_id=audio.orig_path_or_id,
                )
            )
    else:
        batched_audios, sampling_rates, metadatas = batch_audios(audios)

        batched_audios = batched_audios.to(device=torch.device(str(device_type)), dtype=dtype)
        sampling_rate = sampling_rates[0] if isinstance(sampling_rates, List) else sampling_rates
        augmented_audio = augmentation(batched_audios, sample_rate=sampling_rate).samples

        augmented_audio = augmented_audio.detach().cpu()
        return unbatch_audios(augmented_audio, sampling_rates, metadatas)

    return new_audios


augment_audios_pt = pydra.mark.task(augment_audios)



"""This module implements some utilities for evaluating a transcription."""

import jiwer

# TODO: add more metrics which take into account the meaning/intention


def calculate_wer(reference: str, hypothesis: str) -> float:
    """Calculate the Word Error Rate (WER) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The WER score.

    Examples:
        >>> calculate_wer("hello world", "hello duck")
        0.5
    """
    return jiwer.wer(reference, hypothesis)


def calculate_mer(reference: str, hypothesis: str) -> float:
    """Calculate the Match Error Rate (MER) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The MER score.

    Examples:
        >>> calculate_mer("hello world", "hello duck")
        0.5
    """
    return jiwer.mer(reference, hypothesis)


def calculate_wil(reference: str, hypothesis: str) -> float:
    """Calculate the Word Information Lost (WIL) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The WIL score.

    Examples:
        >>> calculate_wil("hello world", "hello duck")
        0.75
    """
    return jiwer.wil(reference, hypothesis)


def calculate_wip(reference: str, hypothesis: str) -> float:
    """Calculate the Word Information Preserved (WIP) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The WIP score.

    Examples:
        >>> calculate_wip("hello world", "hello duck")
        0.25
    """
    return jiwer.wip(reference, hypothesis)


def calculate_cer(reference: str, hypothesis: str) -> float:
    """Calculate the Character Error Rate (CER) between the reference and hypothesis.

    Args:
        reference (str): The ground truth text.
        hypothesis (str): The predicted text.

    Returns:
        float: The CER score.

    Examples:
        >>> calculate_cer("hello world", "hello duck")
        0.45454545454545453
    """
    return jiwer.cer(reference, hypothesis)



"""This module contains functions for voice cloning using KNNVC."""

from typing import Any, Dict, List, Optional

import torch

from senselab.audio.data_structures.audio import Audio
from senselab.utils.data_structures.device import DeviceType, _select_device_and_dtype
from senselab.utils.data_structures.model import TorchModel


class KNNVC:
    """A factory for managing KNNVC pipelines."""

    _pipelines: Dict[str, Any] = {}

    @classmethod
    def _get_knnvc_pipeline(
        cls,
        model: TorchModel,
        prematched_vocoder: bool,
        topk: int,
        device: Optional[DeviceType] = None,
    ) -> Any:  # noqa:ANN401
        """Get or create a KNNVC pipeline.

        Args:
            model (TorchModel): The Torch model to use for the KNNVC pipeline.
            prematched_vocoder (bool): Flag indicating whether to use a pre-matched vocoder.
            topk (int): The number of top matches to consider.
            device (Optional[DeviceType]): The device to run the pipeline on.

        Returns:
            Any: The KNNVC pipeline.
        """
        key = f"{model.path_or_uri}-{model.revision}-{prematched_vocoder}-{topk}-{device}"
        if key not in cls._pipelines:
            device, _ = _select_device_and_dtype(
                user_preference=device, compatible_devices=[DeviceType.CUDA, DeviceType.CPU]
            )
            knn_vc = torch.hub.load(
                model="knn_vc",
                repo_or_dir=model.path_or_uri,
                prematched=prematched_vocoder,
                trust_repo=True,
                pretrained=True,
                device=device.value,
            )
            cls._pipelines[key] = knn_vc
        return cls._pipelines[key]

    @classmethod
    def clone_voices_with_knn_vc(
        cls,
        source_audios: List[Audio],
        target_audios: List[Audio],
        model: TorchModel = TorchModel(path_or_uri="bshall/knn-vc", revision="master"),
        prematched_vocoder: bool = True,
        topk: int = 4,
        device: Optional[DeviceType] = None,
    ) -> List[Audio]:
        """Clone voices from source audios to target audios using KNNVC.

        Args:
            source_audios (List[Audio]): List of source audio objects.
            target_audios (List[Audio]): List of target audio objects.
            model (TorchModel, optional): The Torch model to use for the KNNVC pipeline.
                Defaults to TorchModel(path_or_uri="bshall/knn-vc", revision="master").
            prematched_vocoder (bool, optional): Flag indicating whether to use a pre-matched vocoder. Defaults to True.
            topk (int, optional): The number of top matches to consider. Defaults to 4.
            device (Optional[DeviceType], optional): The device to run the pipeline on. Defaults to None.

        Returns:
            List[Audio]: List of cloned audio objects.

        Raises:
            ValueError: If the audio files are not mono or if the sampling rates are not supported.
        """
        if not isinstance(prematched_vocoder, bool):
            raise TypeError("prematched_vocoder must be a boolean.")

        knn_vc = cls._get_knnvc_pipeline(model=model, prematched_vocoder=prematched_vocoder, topk=topk, device=device)

        cloned_audios = []
        for source_audio, target_audio in zip(source_audios, target_audios):
            if source_audio.waveform.shape[0] > 1 or target_audio.waveform.shape[0] > 1:
                raise ValueError(
                    "Only mono audio files are supported."
                    f"Offending audios: source_audio={source_audio}, target_audio={target_audio}"
                )
            source_sampling_rate = source_audio.sampling_rate
            target_sampling_rate = target_audio.sampling_rate
            # 16kHz is the only supported sampling rate for KNNVC
            supported_sampling_rate = 16000
            if source_sampling_rate != supported_sampling_rate or target_sampling_rate != supported_sampling_rate:
                raise ValueError(
                    f"Sampling rates for the source audio ({source_sampling_rate}) "
                    f"and/or the target audio ({target_sampling_rate}) are not supported."
                    f"Only {supported_sampling_rate} sampling rate is supported."
                    f"Offending audios: source_audio={source_audio}, target_audio={target_audio}"
                )

            source_waveform = source_audio.waveform
            target_waveform = target_audio.waveform

            query_seq = knn_vc.get_features(source_waveform)
            matching_set = knn_vc.get_matching_set([target_waveform])
            out_wav = knn_vc.match(query_seq, matching_set, topk=topk)

            cloned_audios.append(
                Audio(
                    waveform=out_wav,
                    sampling_rate=source_sampling_rate,
                    orig_path_or_id=source_audio.orig_path_or_id,  # TODO: this should be customized
                )
            )

        return cloned_audios



"""Functionality for Pydra."""

from typing import Iterator

import numpy as np
import torch
from pydra.utils.hash import Cache, bytes_repr_sequence_contents, register_serializer


@register_serializer(torch.Tensor)
def bytes_repr_arraylike(obj: torch.Tensor, cache: Cache) -> Iterator[bytes]:
    """Register a serializer for Torch tensors that allows Pydra to properly use them."""
    yield f"{obj.__class__.__module__}{obj.__class__.__name__}:".encode()
    array = np.asanyarray(obj)
    yield f"{array.size}:".encode()
    if array.dtype == "object":
        yield from bytes_repr_sequence_contents(iter(array.ravel()), cache)
    else:
        yield array.tobytes(order="C")


# TODO: Ignore this for now but need to decide how to incorporate Pydra into the package
# Pydra runner
# need function that allows for marking a task (could be obfuscated internally)
# need function that runs the actual code/does the parallelization
# getting results is a bit clunky right now so could clean that up
# internal Pydra wouldn't need obfuscations for add/split but external API might be useful
# perhaps dictionary structure could be intuitive? linked list?

# what if we had a simple split/run structure:
# user: run X task with Y audios (Satra example); perhaps extraneous information of DeviceType, Model to use
#   - kinda easy example of just create the pydra setup ourselves, no user worries
# Senselab "pipelines"/Tasks:

# example:
# RAVDESS audios:
# - need to resample
# - perhaps run loudness normalization
# - run SER
# - run evaluation

# get RAVDESS audios
# split into batches
#   - run batches through resmapling
#   - pass to normalization
#   - run SER
#   - get partial evaluation
# - combine evaluation
# - give average



"""This module provides the Speechbrain interface for speech enhancement."""

import time
from typing import Dict, List, Optional, Tuple

import torch
from speechbrain.inference.separation import SepformerSeparation as separator

from senselab.audio.data_structures.audio import Audio
from senselab.utils.data_structures.device import DeviceType, _select_device_and_dtype
from senselab.utils.data_structures.logging import logger
from senselab.utils.data_structures.model import SpeechBrainModel


class SpeechBrainEnhancer:
    """A factory for managing SpeechBrain enhancement pipelines."""

    _models: Dict[str, separator] = {}

    @classmethod
    def _get_speechbrain_model(
        cls,
        model: SpeechBrainModel,
        device: Optional[DeviceType] = None,
    ) -> Tuple[separator, DeviceType, torch.dtype]:
        """Get or create a SpeechBrain enhancement model.

        Args:
            model (SpeechBrainModel): The SpeechBrain model.
            device (Optional[DeviceType]): The device to run the model on.
                Only CPU and CUDA are supported.

        Returns:
            separator: The SpeechBrain enhancement model.
            device: The device used for the model.
            dtype: The dtype used for the model.
        """
        device, dtype = _select_device_and_dtype(
            user_preference=device, compatible_devices=[DeviceType.CUDA, DeviceType.CPU]
        )
        key = f"{model.path_or_uri}-{model.revision}-{device.value}"
        if key not in cls._models:
            cls._models[key] = separator.from_hparams(source=model.path_or_uri, run_opts={"device": device.value})
        return cls._models[key], device, dtype

    @classmethod
    def enhance_audios_with_speechbrain(
        cls,
        audios: List[Audio],
        model: SpeechBrainModel = SpeechBrainModel(
            path_or_uri="speechbrain/sepformer-wham16k-enhancement", revision="main"
        ),
        device: Optional[DeviceType] = None,
        batch_size: int = 16,
    ) -> List[Audio]:
        """Enhances all audio samples using the given speechbrain model.

        Args:
            audios (List[Audio]): The list of audio objects to be enhanced.
            model (SpeechBrainModel): The SpeechBrain model used for enhancement.
            device (Optional[DeviceType]): The device to run the model on (default is None).
            batch_size (int): The size of batches to use when processing on a GPU.

        Returns:
            List[Audio]: The list of enhanced audio objects.
        """
        # Take the start time of the model initialization
        start_time_model = time.time()
        enhancer, device, _ = cls._get_speechbrain_model(model=model, device=device)
        end_time_model = time.time()
        elapsed_time_model = end_time_model - start_time_model
        logger.info(f"Time taken to initialize the speechbrain model: {elapsed_time_model:.2f} seconds")

        expected_sample_rate = enhancer.hparams.sample_rate

        for audio in audios:
            if audio.waveform.shape[0] != 1:
                raise ValueError(f"Audio waveform must be mono (1 channel), but got {audio.waveform.shape[0]} channels")
            if audio.sampling_rate != expected_sample_rate:
                raise ValueError(
                    "Audio sampling rate "
                    + str(audio.sampling_rate)
                    + " does not match expected "
                    + str(expected_sample_rate)
                )

        # Take the start time of the enhancement
        start_time_enhancement = time.time()
        # Check that all audio objects have the correct sampling rate
        for audio in audios:
            if audio.waveform.shape[0] != 1:
                raise ValueError(f"Audio waveform must be mono (1 channel), but got {audio.waveform.shape[0]} channels")
            if audio.sampling_rate != expected_sample_rate:
                raise ValueError(
                    "Audio sampling rate "
                    + str(audio.sampling_rate)
                    + " does not match expected "
                    + str(expected_sample_rate)
                )

            # Enhance waveforms in a batch
            enhanced_waveform = enhancer.separate_batch(audio.waveform)

            audio.waveform = enhanced_waveform.reshape(1, -1)

        end_time_enhancement = time.time()
        elapsed_time_enhancement = end_time_enhancement - start_time_enhancement
        logger.info(f"Time taken for enhancing the audios: {elapsed_time_enhancement:.2f} seconds")

        return audios



"""This module implements some utilities for the model class."""

import os
from pathlib import Path
from typing import Optional, Union

import requests
import torch
import torchaudio
from huggingface_hub import HfApi
from huggingface_hub.hf_api import ModelInfo
from pydantic import BaseModel, Field, ValidationInfo, field_validator
from typing_extensions import Annotated


class SenselabModel(BaseModel):
    """Base configuration for SenselabModel class."""

    path_or_uri: Union[str, Path]
    revision: Optional[str] = None

    @field_validator("path_or_uri", mode="before")
    def validate_path_or_uri(cls, value: Union[str, Path]) -> Union[str, Path]:
        """Validate the path_or_uri.

        This check is only for files and not for remote resources.
        It does check if the path_or_uri is not empty and if it is an existing file.
        """
        if not value:
            raise ValueError("path_or_uri cannot be empty")

        if isinstance(value, Path) and not os.path.isfile(value):
            raise ValueError("path_or_uri is not an existing file")

        # If the value is a string and looks like an existing file path, convert it to a Path object
        if isinstance(value, str) and os.path.isfile(value):
            value = Path(value)
            if not is_torch_model(value):
                raise ValueError("path_or_uri does not point to a valid torch model")

        return value


class HFModel(SenselabModel):
    """HuggingFace model."""

    revision: Annotated[str, Field(validate_default=True)] = "main"
    info: Optional[ModelInfo] = None

    @field_validator("revision", mode="before")
    def validate_hf_model_id(cls, value: str, info: ValidationInfo) -> Union[str, Path]:
        """Validate the path_or_uri.

        This check is only for remote resources and not for files.
        It checks if the specified Hugging Face model ID and revision exist in the remote Hub.
        """
        path_or_uri = info.data["path_or_uri"]
        if not isinstance(path_or_uri, Path):
            if not check_hf_repo_exists(repo_id=str(path_or_uri), revision=value, repo_type="model"):
                raise ValueError(
                    f"path_or_uri ({path_or_uri}) or specified revision ({value})" " is not a valid Hugging Face model"
                )
        return value

    def get_model_info(self) -> ModelInfo:
        """Gets the model info using the HuggingFace API and saves it as a property."""
        if not self.info:
            api = HfApi()
            self.info = api.model_info(repo_id=self.path_or_uri, revision=self.revision)
        return self.info


class SpeechBrainModel(HFModel):
    """SpeechBrain model."""

    pass


class PyannoteAudioModel(HFModel):
    """PyannoteAudioModel model."""

    pass


class SentenceTransformersModel(HFModel):
    """SentenceTransformersModel model."""

    pass


class TorchModel(SenselabModel):
    """Generic torch model."""

    revision: Annotated[str, Field(validate_default=True)] = "main"

    @field_validator("revision", mode="before")
    def validate_torch_model_id(cls, value: str, info: ValidationInfo) -> Union[str, Path]:
        """Validate the path_or_uri.

        This check is only for remote resources and not for files.
        It checks if the specified torch model ID and revision exist in the remote Hub.
        """
        path_or_uri = info.data["path_or_uri"]
        if not isinstance(path_or_uri, Path):
            if not check_github_repo_exists(repo_id=str(path_or_uri), branch=value):
                raise ValueError("path_or_uri or specified revision is not a valid github repo")
        return value


class TorchAudioModel(SenselabModel):
    """TorchAudio model."""

    revision: Annotated[str, Field(validate_default=True)] = "main"

    @field_validator("revision", mode="before")
    def validate_torchaudio_model_id(cls, value: str, info: ValidationInfo) -> Union[str, Path]:
        """Validate the path_or_uri for torchaudio models.

        This check is only for remote resources and not for files.
        It checks if the specified torchaudio model ID exists.
        """
        path_or_uri = info.data["path_or_uri"]
        if not isinstance(path_or_uri, Path):
            if not check_torchaudio_model_exists(model_id=str(path_or_uri)):
                raise ValueError("path_or_uri is not a valid torchaudio model")
        return value


def check_torchaudio_model_exists(model_id: str) -> bool:
    """Private function to check if a torchaudio model exists."""
    try:
        _ = getattr(torchaudio.pipelines, model_id)
        return True
    except AttributeError:
        return False


def is_torch_model(file_path: Path) -> bool:
    """Check if a file is a torch model."""
    try:
        _ = torch.load(file_path)
        return True
    except Exception:
        return False


def check_hf_repo_exists(repo_id: str, revision: str = "main", repo_type: str = "model") -> bool:
    """Private function to check if a Hugging Face repository exists."""
    api = HfApi()
    try:
        api.list_repo_commits(repo_id=repo_id, revision=revision, repo_type=repo_type)
        return True
    except Exception:
        # raise RuntimeError(f"An error occurred: {e}")
        return False


def check_github_repo_exists(repo_id: str, branch: str = "main") -> bool:
    """Private function to check if a GitHub repository exists."""
    url = f"https://api.github.com/repos/{repo_id}/branches/{branch}"
    response = requests.get(url, timeout=10)
    if response.status_code == 200:
        return True
    elif response.status_code == 404:
        return False
    else:
        response.raise_for_status()
        return False



"""This module implements some utilities to extract speaker embeddings from an audio clip."""

from typing import List, Optional

import torch

from senselab.audio.data_structures.audio import Audio
from senselab.audio.tasks.speaker_embeddings.speechbrain import SpeechBrainEmbeddings
from senselab.utils.data_structures.device import DeviceType
from senselab.utils.data_structures.model import SenselabModel, SpeechBrainModel


def extract_speaker_embeddings_from_audios(
    audios: List[Audio],
    model: SenselabModel = SpeechBrainModel(path_or_uri="speechbrain/spkrec-ecapa-voxceleb", revision="main"),
    device: Optional[DeviceType] = None,
) -> List[torch.Tensor]:
    """Compute the speaker embedding of audio signals.

    Args:
        audios (List[Audio]): A list of Audio objects containing the audio signals and their properties.
        model (SpeechBrainModel): The model used to compute the embeddings
            (default is "speechbrain/spkrec-ecapa-voxceleb").
        device (Optional[DeviceType]): The device to run the model on (default is None).

    Returns:
        List[torch.Tensor]: A list of 1d tensors containing the speaker embeddings for each audio file.

    Raises:
        NotImplementedError: If the model is not a Hugging Face model.

    Examples:
        >>> audios = [Audio.from_filepath("sample.wav")]
        >>> model = SpeechBrainModel(path_or_uri="speechbrain/spkrec-ecapa-voxceleb", revision="main")
        >>> embeddings = extract_speaker_embeddings_from_audios(audios, model, device=DeviceType.CUDA)
        >>> print(embeddings[0].shape)
        torch.Size([192])
    """
    if isinstance(model, SpeechBrainModel):
        return SpeechBrainEmbeddings.extract_speechbrain_speaker_embeddings_from_audios(
            audios=audios, model=model, device=device
        )
    else:
        raise NotImplementedError("The specified model is not supported for now.")



"""Data structures relevant for video tasks and pipelines."""

import os
import uuid
from typing import Dict, List, Optional, Union

import numpy as np
import PIL
import torch
from pydantic import BaseModel, Field, ValidationInfo, field_validator
from torchvision.io import read_video

from senselab.audio.data_structures.audio import Audio
from senselab.utils.constants import SENSELAB_NAMESPACE


class Video(BaseModel):
    """Pydantic model for video and its corresponding metadata.

    Pydantic model for video that holds the necessary attributes, the actual decoded video data
    and the frame rate, to work with videos in python. Contains metadata information as needed
    and has a unique identifier for every video.

    Attributes:
        frames: Represent the video as a Tensor of all of its frames, each of which is an image
            that we represent through a Tensor of (C, H, W)
        frame_rate: Also known as the frames per second (fps), defines the time component
            of a video (often an integer but some use cases of float approximations)
        audio: the audio associated with the Video (optional)
        orig_path_or_id: Optional str for the original path or an ID to track file over time
        metadata: Optional metadata dictionary of information associated with this Video instance
            (e.g. participant demographics, video settings, location information)
    """

    frames: torch.Tensor
    frame_rate: float
    audio: Optional[Audio]
    orig_path_or_id: Optional[str] = Field(default_factory=lambda: str(uuid.uuid4()))
    metadata: Dict = Field(default={})
    model_config = {"arbitrary_types_allowed": True}

    @field_validator("frames", mode="before")
    def check_frames(
        cls, v: Union[torch.Tensor, List[Union[torch.Tensor, np.ndarray, PIL.Image.Image]]], _: ValidationInfo
    ) -> torch.Tensor:
        # print(v)
        """Check that the frames are the correct Tensor shape of (T,H,W,C)."""
        if isinstance(v, torch.Tensor):
            if len(v.shape) != 4:
                raise ValueError(
                    "Expected frames to be of shape (T, H, W, C) where T is the number of frames, \
                                C is the channels, and H and W are the height and width"
                )
            else:
                return v
        elif isinstance(v, List):
            transformed_frames = []
            # print('should get here, is list')
            for frame in v:
                if isinstance(frame, (torch.Tensor, np.ndarray)):
                    if len(frame.shape) != 3:
                        raise ValueError(
                            "Expected frame to be of shape (H, W, C) where \
                                        C is the channels, and H and W are the height and width"
                        )

                    transformed_frames.append(torch.Tensor(frame))
                elif isinstance(frame, PIL.Image.Image):
                    # print('should get here, is image')
                    transformed_frames.append(torch.Tensor(np.array(frame)))
                else:
                    raise ValueError(
                        "Expected each frame in the video to be either a Tensor, numpy array, or PIL.Image"
                    )
            # print(transformed_frames)
            return torch.stack(transformed_frames)
        else:
            raise ValueError("Expected sequence of frames to be either a List of frames or 4d Tensor")

    @classmethod
    def from_filepath(cls, filepath: str, metadata: Dict = {}) -> "Video":
        """Creates a Video instance from a video file.

        Args:
            filepath: Filepath of the video file to read from
            metadata: Additional information associated with the video file
        """
        v_frames, a_frames, v_metadata = read_video(filename=filepath, pts_unit="sec")
        v_fps = v_metadata["video_fps"]
        a_fps = v_metadata["audio_fps"]
        no_ext_filepath = "random_id"  # os.path.splitext(filepath)[0]
        v_audio = Audio(waveform=a_frames, sampling_rate=a_fps, orig_path_or_id=f"{no_ext_filepath}.wav")

        return cls(frames=v_frames, frame_rate=v_fps, audio=v_audio, orig_path_or_id=filepath, metadata=metadata)

    def generate_path(self) -> str:
        """Generate a path like string for this Video.

        Generates a path like string for the Video by either utilizing the orig_path_or_id, checking
        if it is a path (has an extension), otherwise using the id if orig_path_or_id is not an ID
        and giving an extension and relative to the current directory.
        """
        if self.orig_path_or_id:
            if os.path.splitext(self.orig_path_or_id)[-1].lower():
                return self.orig_path_or_id
            else:
                return f"{self.orig_path_or_id}.mp4"
        else:
            return f"{self.id()}.mp4"

    def id(self) -> str:
        """Generate a unique identifier for the Video.

        Generate a unique identifier for the Video where equivalent video frames and frame rate
        and audio generate the same IDs.

        Returns: String UUID of the Video generated by an MD5 hash of the frames and the frame rate and audio
        """
        temp_hash = uuid.uuid3(uuid.uuid3(SENSELAB_NAMESPACE, str(self.frames)), str(self.frame_rate))
        return str(temp_hash) if not self.audio else str(uuid.uuid3(temp_hash, self.audio.id()))

    def __eq__(self, other: object) -> bool:
        """Overloads the default BaseModel equality to correctly check equivalence, ignoring metadata."""
        if isinstance(other, Audio):
            return self.id() == other.id()
        return False



"""This module defines APIs for the user to run Speech Emotion Recognition (SER) on sets of audios."""

import warnings
from typing import Dict, List, Optional, Tuple

from transformers import AutoConfig, pipeline

from senselab.audio.data_structures.audio import Audio
from senselab.utils.data_structures.device import DeviceType, _select_device_and_dtype
from senselab.utils.data_structures.model import HFModel


def audio_classification_with_hf_models(
    audios: List[Audio], model: HFModel, device: Optional[DeviceType] = None
) -> List[List[Dict]]:
    """General audio classification functionality utilitzing HuggingFace pipelines.

    Classifies all audios, with no underlying assumptions on what the classification labels are,
    and returns the output that the pipeline gives.

    Args:
        audios: List of Audio objects that we want to run classification on
        model: The HuggingFace model that will be used for running the inference
        device: The device to run inference on

    Returns:
        List of Lists of Dictionaries where each corresponds to the audio that it was ran on and the List of
            Dictionaries are of the form [{'label': 'some_label', 'score': some_value},...]

    Raises:
        ValueError if the given model does not have the audio-classification pipeline tag
        UserWarning if the model tags don't include endpoints_compatible (seen on HuggingFace as Inference Endpoints)
            as the behavior of the model might not output as expected as a result.
    """
    model_info = model.get_model_info()

    if "audio-classification" not in model_info.pipeline_tag:
        raise ValueError(f"The model '{model.path_or_uri}' is not suitable for audio classification. SORRY!")
    if "endpoints_compatible" not in model_info.tags:
        warnings.warn(
            UserWarning(f"The model '{model.path_or_uri}' has not been tagged as an Inference Endpoint and \
                                  so we cannot guarantee its input and outputs are as expected")
        )

    device, _ = _select_device_and_dtype(user_preference=device, compatible_devices=[DeviceType.CUDA, DeviceType.CPU])

    classification_pipeline = pipeline(
        task="audio-classification",
        model=model.path_or_uri,
        revision=model.revision,
        device=device.value,
    )
    classification_outputs = []

    # TODO: figure out adding batching and GPU support
    for audio in audios:
        classification_outputs.append(classification_pipeline(audio.waveform.numpy().squeeze()))

    return classification_outputs


def speech_emotion_recognition_with_hf_models(
    audios: List[Audio], model: HFModel, device: Optional[DeviceType] = None
) -> List[Tuple[str, Dict]]:
    """Function for running speech emotion recognition tasks using HuggingFace models.

    Uses an audio classification pipeline to run speech emotion recognition for every audio in audios using the
    specified model. For each audio, we return the emotion that the model suggested has the highest probability
    for single label classification as well as the full emotion output for multi-label classification. Function can
    also be used for continuous speech emotion recognition, where the second element of each tuple will contain a
    dictionary of the predicted continuous values.

    Args:
        audios: List of Audio objects that we want to run speech emotion recognition on. If you wish to get the
            emotion at different segments of the audio (e.g. when different speaker talk) please run the audios
            through an appropriate segmentation task.
        model: The HuggingFace model that will be used for running the inference
        device: The device to run inference on

    Returns:
        List of tuples where the first value is the single label classification of the audio (if appropriate) and the
            second value is the full model output, which might be the probabilities for each evaluated emotion
            (for discrete SER) or the continuous emotion predictions.

    Raises:
        ValueError if the given model is not properly tagged with 'speech-emotion-recognition' or 'emotion-recognition'
            or the model configuration does not contain an id2label property that predicts commonly used
            emotions (happy, sad, neutral, positive, negative, etc.)
    """
    model_info = model.get_model_info()

    tags = model_info.tags

    if not ("speech-emotion-recognition" in tags or "emotion-recognition" in tags) and not _are_emotions_in_config(
        model
    ):
        raise ValueError(
            f"The model '{model.path_or_uri}' is not suitable for speech emotion recognition. Please +"
            "validate that it has the correct tags or use the more generic "
            + "'audio_classification_with_hf_models' function."
        )

    audio_classifications = audio_classification_with_hf_models(audios, model, device)
    # print(audio_classifications)
    ser_output = []
    for classification in audio_classifications:
        classification_output = {}
        for label_score in classification:
            classification_output[label_score["label"]] = label_score["score"]
        single_classification = max(classification_output, key=lambda x: classification_output[x])
        ser_output.append((single_classification, classification_output))
    return ser_output


def _are_emotions_in_config(model: HFModel) -> bool:
    config = AutoConfig.from_pretrained(model.path_or_uri)
    id2label = config.id2label
    # print(id2label)
    if id2label:
        # id2label = model.model_info.config['id2label']
        labels = list(id2label.values())
        if "positive" in labels and "negative" in labels and "neutral" in labels:  # Simple, discrete
            return True
        if "valence" in labels or "arousal" in labels or "dominance" in labels:  # conitnuous emotions
            return True
        if (
            "happy" in labels
            or "happiness" in labels
            or "sad" in labels
            or "sadness" in labels
            or "fear" in labels
            or "anger" in labels
            or "disgust" in labels
        ):
            return True
    return False



"""This module contains functions for computing the normalized cross-correlation between two signals."""

import numpy as np
import pydra
import torch
from scipy.signal import correlate


def compute_normalized_cross_correlation(signal1: torch.Tensor, signal2: torch.Tensor) -> torch.Tensor:
    """Calculate the normalized cross-correlation between two signals.

    Args:
        signal1 (torch.Tensor): The first input signal as a PyTorch tensor.
        signal2 (torch.Tensor): The second input signal as a PyTorch tensor.

    Returns:
        torch.Tensor: The normalized cross-correlation value between the two input signals.

    Examples:
        >>> signal1 = torch.tensor([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> signal2 = torch.tensor([2.0, 3.0, 4.0])
        >>> normalized_cross_correlation(signal1, signal2)
        Tensor([0.30151134, 0.51298918, 0.77459667, 0.9486833 , 0.90453403, 0.70710678, 0.43643578])

    Note:
        This function assumes the input signals are one-dimensional
        and contain sufficient elements for meaningful cross-correlation.
    """
    # Ensure the inputs are 1D tensors
    if signal1.ndim != 1 or signal2.ndim != 1:
        raise ValueError("Input signals must be one-dimensional")

    # Convert PyTorch tensors to NumPy arrays
    signal1 = signal1.numpy()
    signal2 = signal2.numpy()

    # Calculate the energy of each signal
    energy_signal1 = np.sum(signal1**2)
    energy_signal2 = np.sum(signal2**2)

    # Check for zero energy to avoid division by zero
    if energy_signal1 == 0 or energy_signal2 == 0:
        raise ZeroDivisionError("One of the input signals has zero energy, causing division by zero in normalization")

    # Compute the cross-correlation
    cross_correlation = correlate(signal1, signal2)

    # Calculate the normalized cross-correlation
    normalized_cross_correlation = cross_correlation / np.sqrt(energy_signal1 * energy_signal2)

    print(normalized_cross_correlation)
    return torch.Tensor(normalized_cross_correlation)


compute_normalized_cross_correlation_pt = pydra.mark.task(compute_normalized_cross_correlation)



"""This module contains functions for extracting features from pre-trained self-supervised models."""

from typing import Dict, List, Optional

import torch
from transformers import AutoModel, AutoTokenizer

from senselab.utils.data_structures.device import DeviceType, _select_device_and_dtype
from senselab.utils.data_structures.model import HFModel


class HFFactory:
    """A factory for managing self-supervised models from Hugging Face."""

    _tokenizers: Dict[str, AutoTokenizer] = {}
    _models: Dict[str, AutoModel] = {}

    @classmethod
    def _get_tokenizer(
        cls,
        model: HFModel,
    ) -> AutoTokenizer:
        """Get or create a tokenizer for SSL model.

        Args:
            model (HFModel): The HuggingFace model.

        Returns:
            AutoTokenizer: The tokenizer for the model.
        """
        key = f"{model.path_or_uri}-{model.revision}"
        if key not in cls._tokenizers:
            cls._tokenizers[key] = AutoTokenizer.from_pretrained(
                pretrained_model_name_or_path=model.path_or_uri,
                revision=model.revision,
            )
        return cls._tokenizers[key]

    @classmethod
    def _load_model(
        cls,
        model: HFModel,
        device: DeviceType,
    ) -> AutoModel:
        """Load weights of SSL model.

        Args:
            model (HFModel): The Hugging Face model.
            device (DeviceType): The device to run the model on.

        Returns:
            AutoModel: The SSL model.
        """
        key = f"{model.path_or_uri}-{model.revision}-{device.value}"
        if key not in cls._models:
            cls._models[key] = AutoModel.from_pretrained(model.path_or_uri, revision=model.revision).to(device.value)
        return cls._models[key]

    @classmethod
    def extract_text_embeddings(
        cls,
        pieces_of_text: List[str],
        model: HFModel = HFModel(path_or_uri="sentence-transformers/all-MiniLM-L6-v2"),
        device: Optional[DeviceType] = None,
    ) -> List[torch.Tensor]:
        """Extracts embeddings from a list of strings using a Hugging Face model.

        Args:
            pieces_of_text (List[str]): A list of strings to extract embeddings from.
            model (HFModel, optional): A Hugging Face model configuration.
                Defaults to HFModel(path_or_uri="sentence-transformers/all-MiniLM-L6-v2").
            device (Optional[DeviceType], optional): The device to run the model on.
                Defaults to None.

        Returns:
            List[torch.Tensor]: A list of embeddings for the input strings.
        """
        device, _ = _select_device_and_dtype(
            user_preference=device, compatible_devices=[DeviceType.CUDA, DeviceType.CPU]
        )
        # Load tokenizer and model
        tokenizer = cls._get_tokenizer(model=model)
        ssl_model = cls._load_model(model=model, device=device)

        embeddings = []

        # Process each piece of text individually
        for text in pieces_of_text:
            # Tokenize sentence
            encoded_input = tokenizer(text, return_tensors="pt").to(device)

            # Compute token embeddings
            with torch.no_grad():
                model_output = ssl_model(**encoded_input, output_hidden_states=True)
                hidden_states = model_output.hidden_states
                concatenated_hidden_states = torch.cat([state.unsqueeze(0) for state in hidden_states], dim=0)
                embeddings.append(concatenated_hidden_states.squeeze())

        return embeddings



"""This module contains functions for extracting openSMILE features."""

from typing import Any, Dict, List

import opensmile
import pydra

from senselab.audio.data_structures.audio import Audio


class OpenSmileFeatureExtractorFactory:
    """A factory for managing openSMILE feature extractors."""

    _extractors: Dict[str, opensmile.Smile] = {}

    @classmethod
    def get_opensmile_extractor(cls, feature_set: str, feature_level: str) -> opensmile.Smile:
        """Get or create an openSMILE feature extractor.

        Args:
            feature_set (str): The openSMILE feature set.
            feature_level (str): The openSMILE feature level.

        Returns:
            opensmile.Smile: The openSMILE feature extractor.
        """
        key = f"{feature_set}-{feature_level}"
        if key not in cls._extractors:
            cls._extractors[key] = opensmile.Smile(
                feature_set=opensmile.FeatureSet[feature_set],
                feature_level=opensmile.FeatureLevel[feature_level],
            )
        return cls._extractors[key]


def extract_opensmile_features_from_audios(
    audios: List[Audio],
    feature_set: str = "eGeMAPSv02",
    feature_level: str = "Functionals",
) -> List[Dict[str, Any]]:
    """Apply feature extraction across a list of audio files.

    Args:
        audios (List[Audio]): The list of audio objects to extract features from.
        feature_set (str): The openSMILE feature set (default is "eGeMAPSv02").
        feature_level (str): The openSMILE feature level (default is "Functionals").

    Returns:
        List[Dict[str, Any]]: The list of feature dictionaries for each audio.
    """

    def _extract_feats_from_audio(sample: Audio, smile: opensmile.Smile) -> Dict[str, Any]:
        """Extract features from a single audio sample using openSMILE.

        Args:
            sample (Audio): The audio object.
            smile (opensmile.Smile): The openSMILE feature extractor.

        Returns:
            Dict[str, Any]: The extracted features as a dictionary.
        """
        audio_array = sample.waveform.squeeze().numpy()
        sampling_rate = sample.sampling_rate
        sample_features = smile.process_signal(audio_array, sampling_rate)
        # Convert to a dictionary with float values and return it
        return {
            k: v[0] if isinstance(v, list) and len(v) == 1 else v for k, v in sample_features.to_dict("list").items()
        }

    smile = OpenSmileFeatureExtractorFactory.get_opensmile_extractor(feature_set, feature_level)
    features = [_extract_feats_from_audio(audio, smile) for audio in audios]
    return features


extract_opensmile_features_from_audios_pt = pydra.mark.task(extract_opensmile_features_from_audios)



"""This module provides the implementation of torchaudio squim utilities for audio features extraction."""

from typing import Any, Dict, List

import pydra
from torchaudio.pipelines import SQUIM_OBJECTIVE, SQUIM_SUBJECTIVE

from senselab.audio.data_structures.audio import Audio

objective_model = SQUIM_OBJECTIVE.get_model()
subjective_model = SQUIM_SUBJECTIVE.get_model()


def extract_objective_quality_features_from_audios(audio_list: List[Audio]) -> Dict[str, Any]:
    """Extracts objective audio features from a list of Audio objects.

    Currently, Torchaudio-Squim model only supports mono audio at 16000 Hz sampling rate.

    Args:
        audio_list (List[Audio]): List of Audio objects.

    Returns:
        Dict[str, Any]: Dictionary containing extracted features.
    """
    if any(audio.waveform.shape[0] != 1 for audio in audio_list):
        raise ValueError("Only mono audio is supported by Torchaudio-Squim model.")

    if any(audio.sampling_rate != 16000 for audio in audio_list):
        raise ValueError("Only 16000 Hz sampling rate is supported by Torchaudio-Squim model.")

    features: Dict[str, Any] = {"stoi": [], "pesq": [], "si_sdr": []}

    for audio in audio_list:
        stoi, pesq, si_sdr = objective_model(audio.waveform)
        features["stoi"].append(stoi.item())
        features["pesq"].append(pesq.item())
        features["si_sdr"].append(si_sdr.item())

    return features


def extract_subjective_quality_features_from_audios(
    audio_list: List[Audio], non_matching_references: List[Audio]
) -> Dict[str, Any]:
    """Extracts subjective audio features from a list of Audio objects.

    Currently, Torchaudio-Squim model only supports mono audio at 16000 Hz sampling rate.

    Args:
        audio_list (List[Audio]): List of Audio objects.
        non_matching_references (List[Audio]): Reference Audio objects for the subjective model.

    Returns:
        Dict[str, Any]: Dictionary containing extracted features.
    """
    # Check if any audio is not mono
    if any(audio.waveform.shape[0] != 1 for audio in audio_list) or any(
        ref.waveform.shape[0] != 1 for ref in non_matching_references
    ):
        raise ValueError("Only mono audio is supported by Torchaudio-Squim model.")

    # Check if any audio has a sampling rate other than 16000 Hz
    if any(audio.sampling_rate != 16000 for audio in audio_list) or any(
        ref.sampling_rate != 16000 for ref in non_matching_references
    ):
        raise ValueError("Only 16000 Hz sampling rate is supported by Torchaudio-Squim model.")

    features: Dict[str, Any] = {"mos": []}

    for i, audio in enumerate(audio_list):
        mos = subjective_model(audio.waveform, non_matching_references[i].waveform)
        features["mos"].append(mos.item())

    return features


extract_objective_quality_features_from_audios_pt = pydra.mark.task(extract_objective_quality_features_from_audios)
extract_subjective_quality_features_from_audios_pt = pydra.mark.task(extract_subjective_quality_features_from_audios)



"""This module implements some utilities for the text-to-speech task."""

from typing import List, Optional

from senselab.audio.data_structures.audio import Audio
from senselab.audio.tasks.text_to_speech.huggingface import HuggingFaceTTS
from senselab.utils.data_structures.device import DeviceType
from senselab.utils.data_structures.language import Language
from senselab.utils.data_structures.model import HFModel, SenselabModel


def synthesize_texts(
    texts: List[str],
    model: SenselabModel = HFModel(path_or_uri="suno/bark", revision="main"),
    language: Optional[Language] = None,
    device: Optional[DeviceType] = None,
) -> List[Audio]:
    """Synthesizes speech from all texts using the given model.

    Args:
        texts (List[str]): The list of text strings to be synthesized.
        model (SenselabModel): The model used for synthesis
            (Default is "suno/bark").
        language (Optional[Language]): The language of the text (default is None).
        device (Optional[DeviceType]): The device to run the model on (default is None).

    Returns:
        List[Audio]: The list of synthesized audio objects.

    Todo:
        - Include more models
        - Include voice cloning for TTS
    """
    if isinstance(model, HFModel):
        return HuggingFaceTTS.synthesize_texts_with_transformers(texts=texts, model=model, device=device)
    else:
        raise NotImplementedError("Only Hugging Face models are supported for now.")



"""This module contains functions for extracting features from pre-trained self-supervised models."""

from typing import Dict, List, Optional

import torch
from transformers import AutoFeatureExtractor, AutoModel

from senselab.audio.data_structures.audio import Audio
from senselab.utils.data_structures.device import DeviceType, _select_device_and_dtype
from senselab.utils.data_structures.model import HFModel


class SSLEmbeddingsFactory:
    """A factory for managing self-supervised models."""

    _feat_extractor: Dict[str, AutoFeatureExtractor] = {}
    _model: Dict[str, AutoModel] = {}

    @classmethod
    def _get_feature_extractor(
        cls,
        model: HFModel,
        cache_dir: str = "~/",
    ) -> AutoFeatureExtractor:
        """Get or create a feature extractor for SSL model.

        Args:
            model (HFModel): The HuggingFace model.
            cache_dir (str): The path to where the model's weights will be saved.

        Returns:
            AutoFeatureExtractor: The feature extractor for the model.
        """
        key = f"{model.path_or_uri}-{model.revision}"
        if key not in cls._feat_extractor:
            cls._feat_extractor[key] = AutoFeatureExtractor.from_pretrained(
                model.path_or_uri, revision=model.revision, cache_dir=cache_dir
            )
        return cls._feat_extractor[key]

    @classmethod
    def _load_model(
        cls,
        model: HFModel,
        device: DeviceType,
        cache_dir: str = "~/",
    ) -> AutoModel:
        """Load weights of SSL model.

        Args:
            model (HFModel): The Hugging Face model.
            cache_dir (str): The path to where the model's weights will be saved.
            device (DeviceType): The device to run the model on.

        Returns:
            AutoModel: The SSL model.
        """
        key = f"{model.path_or_uri}-{model.revision}-{device.value}"
        if key not in cls._model:
            cls._model[key] = AutoModel.from_pretrained(
                model.path_or_uri, revision=model.revision, cache_dir=cache_dir
            ).to(device.value)
        return cls._model[key]

    @classmethod
    def extract_ssl_embeddings(
        cls,
        audios: List[Audio],
        model: HFModel,
        cache_dir: str = "~/",
        device: Optional[DeviceType] = None,
    ) -> List[torch.Tensor]:
        """Compute the ssl embeddings of audio signals.

        Args:
            audios (List[Audio]): A list of Audio objects containing the audio signals and their properties.
            model (HFModel): The model used to compute the embeddings
                (default is "facebook/wav2vec2-base").
            cache_dir (str): The path to where the model's weights will be saved.
            device (Optional[DeviceType]): The device to run the model on (default is None).
                Only CPU and CUDA are supported.

        Returns:
            List[torch.Tensor]: A list of tensors containing the ssl embeddings for each audio file.
        """
        device, _ = _select_device_and_dtype(
            user_preference=device, compatible_devices=[DeviceType.CUDA, DeviceType.CPU]
        )
        # Load feature extractor and model
        feat_extractor = cls._get_feature_extractor(model=model, cache_dir=cache_dir)
        sampling_rate = feat_extractor.sampling_rate
        ssl_model = cls._load_model(model=model, cache_dir=cache_dir, device=device)

        # Check that all audio objects have the correct sampling rate
        for audio in audios:
            if audio.waveform.shape[0] != 1:
                raise ValueError(f"Audio waveform must be mono (1 channel), but got {audio.waveform.shape[0]} channels")
            if audio.sampling_rate != sampling_rate:
                raise ValueError(
                    "Audio sampling rate " + str(audio.sampling_rate) + " does not match expected " + str(sampling_rate)
                )
        # Pre-process audio using the SSL mode feature extractor
        preprocessed_audios = [
            feat_extractor(audio.waveform, sampling_rate=sampling_rate, return_tensors="pt") for audio in audios
        ]

        # Extract embeddings (hidden states from all layers) from pre-trained model
        embeddings = [
            ssl_model(audio.input_values.squeeze(0).to(device.value), output_hidden_states=True)
            for audio in preprocessed_audios
        ]
        embeddings = [torch.cat(embedding.hidden_states) for embedding in embeddings]

        return embeddings



"""This module defines a pydra API for the video input and output task."""

import pydra

from senselab.video.tasks.input_output import extract_audios_from_local_videos

extract_audios_from_local_videos_pt = pydra.mark.task(extract_audios_from_local_videos)



"""Align function based on WhisperX implementation."""

import math
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
import torch
from nltk.tokenize.punkt import PunktParameters, PunktSentenceTokenizer
from threadpoolctl import threadpool_limits
from transformers import Wav2Vec2ForCTC, Wav2Vec2Processor

from senselab.audio.data_structures.audio import Audio
from senselab.audio.tasks.forced_alignment.constants import (
    DEFAULT_ALIGN_MODELS_HF,
    LANGUAGES_WITHOUT_SPACES,
    MINIMUM_SEGMENT_SIZE,
    PUNKT_ABBREVIATIONS,
    SAMPLE_RATE,
)
from senselab.audio.tasks.forced_alignment.data_structures import (
    AlignedTranscriptionResult,
    Point,
    Segment,
    SingleAlignedSegment,
    SingleSegment,
    SingleWordSegment,
)
from senselab.audio.tasks.preprocessing.preprocessing import extract_segments, pad_audios
from senselab.utils.data_structures.device import DeviceType, _select_device_and_dtype
from senselab.utils.data_structures.language import Language
from senselab.utils.data_structures.model import HFModel
from senselab.utils.data_structures.script_line import ScriptLine


def _preprocess_segments(
    transcript: List[SingleSegment],
    model_dictionary: Dict[str, int],
    model_lang: Language,
    print_progress: bool,
    combined_progress: bool,
) -> List[SingleSegment]:
    """Preprocess transcription segments by filtering characters, handling spaces, and preparing text.

    Args:
        transcript (List[SingleSegment]): The list of transcription segments.
        model_dictionary (Dict[str, int]): Dictionary for the alignment model.
        model_lang (Language): Language of the model.
        print_progress (bool): Whether to print progress.
        combined_progress (bool): Whether to combine progress percentage.

    Returns:
        List[SingleSegment]: The preprocessed transcription segments.
    """
    total_segments = len(transcript)

    for sdx, segment in enumerate(transcript):
        if print_progress:
            base_progress = ((sdx + 1) / total_segments) * 100
            percent_complete = (50 + base_progress / 2) if combined_progress else base_progress
            print(f"Progress: {percent_complete:.2f}%...")

        num_leading = len(segment["text"]) - len(segment["text"].lstrip())
        num_trailing = len(segment["text"]) - len(segment["text"].rstrip())
        text = segment["text"]

        # Split into words
        if model_lang.alpha_2 not in LANGUAGES_WITHOUT_SPACES:
            per_word = text.split(" ")
        else:
            per_word = [text]

        clean_char, clean_cdx = [], []
        for cdx, char in enumerate(text):
            char_ = char.lower()
            if model_lang.alpha_2 not in LANGUAGES_WITHOUT_SPACES:
                char_ = char_.replace(" ", "|")

            if cdx < num_leading or cdx > len(text) - num_trailing - 1:
                continue
            elif char_ in model_dictionary.keys():
                clean_char.append(char_)
                clean_cdx.append(cdx)

        clean_wdx = []
        for wdx, wrd in enumerate(per_word):
            if any(c in model_dictionary.keys() for c in wrd):
                clean_wdx.append(wdx)

        punkt_param = PunktParameters()
        punkt_param.abbrev_types = set(PUNKT_ABBREVIATIONS)
        sentence_splitter = PunktSentenceTokenizer(punkt_param)
        sentence_spans = list(sentence_splitter.span_tokenize(text))

        segment["clean_char"] = clean_char
        segment["clean_cdx"] = clean_cdx
        segment["clean_wdx"] = clean_wdx
        segment["sentence_spans"] = sentence_spans

    return transcript


def _can_align_segment(
    segment: SingleSegment, model_dictionary: Dict[str, int], t1: float, max_duration: float
) -> bool:
    """Checks if a segment can be aligned.

    Args:
        segment (SingleSegment): The segment to check.
        model_dictionary (Dict[str, int]): Dictionary for character indices.
        t1 (float): Start time of the segment.
        max_duration (float): Maximum duration of the audio.

    Returns:
        bool: True if the segment can be aligned, False otherwise.
    """
    if segment["clean_char"] is None or len(segment["clean_char"]) == 0:
        return False
    if t1 >= max_duration:
        return False
    return True


def _get_prediction_matrix(
    model: torch.nn.Module,
    waveform_segment: torch.Tensor,
    lengths: Optional[torch.Tensor],
    model_type: str,
    device: DeviceType,
) -> torch.Tensor:
    """Generate prediction matrix from the alignment model.

    Args:
        model (torch.nn.Module): The alignment model.
        waveform_segment (torch.Tensor): The audio segment to be processed.
        lengths (Optional[torch.Tensor]): Lengths of the audio segments.
        model_type (str): The type of the model ('torchaudio' or 'huggingface').
        device (DeviceType): The device to run the model on.

    Returns:
        torch.Tensor: The prediction matrix.
    """
    with torch.inference_mode():
        if model_type == "torchaudio":
            emissions, _ = model(waveform_segment.to(device.value), lengths=lengths)
        elif model_type == "huggingface":
            emissions = model(waveform_segment.to(device.value)).logits
        else:
            raise NotImplementedError(f"Align model of type {model_type} not supported.")

        emissions = torch.log_softmax(emissions, dim=-1)

    return emissions


def _assign_timestamps_to_characters(
    text: str, segment: SingleSegment, char_segments: list, ratio: float, t1: float, model_lang: Language
) -> pd.DataFrame:
    """Assigns timestamps to aligned characters and organizes them into a DataFrame.

    Args:
        text (str): The text to align with the segment.
        segment (SingleSegment): The segment containing character indices.
        char_segments (list): List of character segments with alignment information.
        ratio (float): The ratio of duration to waveform segment size.
        t1 (float): Start time of the segment.
        model_lang (Language): Language of the model.

    Returns:
        pd.DataFrame: DataFrame containing character alignments with timestamps and word indices.
    """
    char_segments_arr = []
    word_idx = 0
    for cdx, char in enumerate(text):
        start, end, score = None, None, None
        if segment["clean_cdx"] is not None and cdx in segment["clean_cdx"]:
            char_seg = char_segments[segment["clean_cdx"].index(cdx)]
            start = round(char_seg.start * ratio + t1, 3)
            end = round(char_seg.end * ratio + t1, 3)
            score = round(char_seg.score, 3)

        char_segments_arr.append(
            {
                "char": char,
                "start": start,
                "end": end,
                "score": score,
                "word-idx": word_idx,
            }
        )

        if model_lang.alpha_2 in LANGUAGES_WITHOUT_SPACES:
            word_idx += 1
        elif cdx == len(text) - 1 or text[cdx + 1] == " ":
            word_idx += 1

    return pd.DataFrame(char_segments_arr)


def _align_subsegments(
    segment: SingleSegment,
    char_segments_df: pd.DataFrame,
    text: str,
    word_segments: List[SingleWordSegment],
    aligned_subsegments: List[SingleAlignedSegment],
    return_char_alignments: bool,
) -> None:
    """Aligns sentence spans to create subsegments and update word segments.

    Args:
        segment (SingleSegment): The segment containing sentence spans.
        char_segments_df (pd.DataFrame): DataFrame with character alignments.
        text (str): The text to align with the segment.
        word_segments (List[SingleWordSegment]): List to store word segments.
        aligned_subsegments (List[SingleAlignedSegment]): List to store aligned subsegments.
        return_char_alignments (bool): Flag to return character alignments.

    Returns:
        None: The function modifies the word_segments and aligned_subsegments lists in place.
    """
    for sdx, (sstart, send) in enumerate(segment["sentence_spans"] or []):
        curr_chars = char_segments_df.loc[(char_segments_df.index >= sstart) & (char_segments_df.index <= send)]
        char_segments_df.loc[(char_segments_df.index >= sstart) & (char_segments_df.index <= send), "sentence-idx"] = (
            sdx
        )

        sentence_text = text[sstart:send]
        sentence_start = curr_chars["start"].min()
        end_chars = curr_chars[curr_chars["char"] != " "]
        sentence_end = end_chars["end"].max()
        sentence_words = []

        for word_idx in curr_chars["word-idx"].unique():
            word_chars = curr_chars.loc[curr_chars["word-idx"] == word_idx]
            word_text = "".join(word_chars["char"].tolist()).strip()
            if len(word_text) == 0:
                continue

            word_chars = word_chars[word_chars["char"] != " "]

            word_start = word_chars["start"].min()
            word_end = word_chars["end"].max()
            word_score = round(word_chars["score"].mean(), 3)

            word_segment = SingleWordSegment(word=word_text, start=word_start, end=word_end, score=word_score)

            sentence_words.append(word_segment)
            word_segments.append(word_segment)

        aligned_subsegment = SingleAlignedSegment(
            text=sentence_text, start=sentence_start, end=sentence_end, words=sentence_words, chars=word_chars
        )
        aligned_subsegments.append(aligned_subsegment)

        if return_char_alignments:
            curr_chars = curr_chars[["char", "start", "end", "score"]]
            curr_chars.fillna(-1, inplace=True)
            curr_chars = curr_chars.to_dict("records")
            curr_chars = [{key: val for key, val in char.items() if val != -1} for char in curr_chars]
            aligned_subsegments[-1]["chars"] = curr_chars


def _align_single_segment(
    segment: SingleSegment,
    model: torch.nn.Module,
    model_dictionary: Dict[str, int],
    model_lang: Language,
    model_type: str,
    audio: Audio,
    device: DeviceType,
    t1: float,
    t2: float,
    return_char_alignments: bool,
    interpolate_method: str,
    aligned_segment: SingleAlignedSegment,
    aligned_segments: List[SingleAlignedSegment],
    word_segments: List[SingleWordSegment],
) -> None:
    """Processes and aligns a single segment.

    Args:
        segment (SingleSegment): The segment to align.
        model (torch.nn.Module): The alignment model.
        model_dictionary (Dict[str, int]): Dictionary for character indices.
        model_lang (Language): Language of the model.
        model_type (str): Either 'huggingface' or 'torchaudio'.
        audio (Audio): The audio data.
        device (DeviceType): The device to run the model on.
        t1 (float): Start time of the segment.
        t2 (float): End time of the segment.
        return_char_alignments (bool): Flag to return character alignments.
        interpolate_method (str): Method for interpolating NaNs.
        aligned_segment (SingleAlignedSegment ): The aligned segment data.
        aligned_segments (List[SingleAlignedSegment]): List to store aligned segments.
        word_segments (List[SingleWordSegment]): List to store word segments.

    Returns:
        None: The function modifies the aligned_segments and word_segments lists in place.
    """
    text_clean = "".join(segment["clean_char"] or [])
    tokens = [model_dictionary[c] for c in text_clean]

    extracted_segment = extract_segments([(audio, [(t1, t2)])])[0][0]
    lengths = extracted_segment.waveform.shape[-1]
    waveform_segment = pad_audios([extracted_segment], MINIMUM_SEGMENT_SIZE)[0].waveform

    emissions = _get_prediction_matrix(model, waveform_segment, lengths, model_type, device)
    emission = emissions[0].cpu().detach()

    blank_id = 0
    for char, code in model_dictionary.items():
        if char == "[pad]" or char == "<pad>":
            blank_id = code

    trellis = _get_trellis(emission, tokens, blank_id)
    path = _backtrack(trellis, emission, tokens, blank_id)

    if path is None:
        print(f'Failed to align segment ("{segment["text"]}"): backtrack failed, resorting to original...')
        aligned_segments.append(aligned_segment)
        return

    char_segments = _merge_repeats(path, text_clean)

    duration = t2 - t1
    ratio = duration * waveform_segment.size(0) / (trellis.size(0) - 1)

    char_segments_df = _assign_timestamps_to_characters(segment["text"], segment, char_segments, ratio, t1, model_lang)

    aligned_subsegments: List[SingleAlignedSegment] = []
    if isinstance(char_segments_df, pd.DataFrame):
        char_segments_df["sentence-idx"] = None
    else:
        raise TypeError("char_segments_df must be a pandas DataFrame.")

    if segment["sentence_spans"] is not None:
        _align_subsegments(
            segment=segment,
            char_segments_df=char_segments_df,
            text=segment["text"],
            word_segments=word_segments,
            aligned_subsegments=aligned_subsegments,
            return_char_alignments=True,
        )

        if aligned_subsegments:
            aligned_subsegments_df = pd.DataFrame(aligned_subsegments)

            aligned_subsegments_df["start"] = _interpolate_nans(
                aligned_subsegments_df["start"], method=interpolate_method
            )
            aligned_subsegments_df["end"] = _interpolate_nans(aligned_subsegments_df["end"], method=interpolate_method)
            agg_dict = {"text": " ".join, "words": "sum"}
            if model_lang.alpha_2 in LANGUAGES_WITHOUT_SPACES:
                agg_dict["text"] = "".join
            if return_char_alignments:
                agg_dict["chars"] = "sum"
            aligned_subsegments_df.groupby(["start", "end"], as_index=False).agg(agg_dict)
            aligned_subsegments = aligned_subsegments_df.to_dict("records")

    aligned_segments.extend(aligned_subsegments)


def _get_trellis(emission: torch.Tensor, tokens: List[int], blank_id: int = 0) -> torch.Tensor:
    """Gets the trellis for token alignment.

    Args:
        emission (torch.Tensor): The emission matrix from the model.
        tokens (List[int]): The token IDs.
        blank_id (int): The ID for the blank token.

    Returns:
        torch.Tensor: The trellis matrix.
    """
    num_frame = emission.size(0)
    num_tokens = len(tokens)

    trellis = torch.empty((num_frame + 1, num_tokens + 1))
    trellis[0, 0] = 0
    trellis[1:, 0] = torch.cumsum(emission[:, 0], 0)
    trellis[0, -num_tokens:] = -float("inf")
    trellis[-num_tokens:, 0] = float("inf")

    for t in range(num_frame):
        trellis[t + 1, 1:] = torch.maximum(
            trellis[t, 1:] + emission[t, blank_id],
            trellis[t, :-1] + emission[t, tokens],
        )
    return trellis


def _backtrack(
    trellis: torch.Tensor, emission: torch.Tensor, tokens: List[int], blank_id: int = 0
) -> Optional[List[Point]]:
    """Backtracks to find the best path through the trellis.

    Args:
        trellis (torch.Tensor): The trellis matrix.
        emission (torch.Tensor): The emission matrix from the model.
        tokens (List[int]): The token IDs.
        blank_id (int): The ID for the blank token.

    Returns:
        Optional[List[Point]]: The best path as a list of Points.
    """
    j = trellis.size(1) - 1
    t_start = int(torch.argmax(trellis[:, j]).item())

    path = []
    for t in range(t_start, 0, -1):
        stayed = trellis[t - 1, j] + emission[t - 1, blank_id]
        changed = trellis[t - 1, j - 1] + emission[t - 1, tokens[j - 1]]
        prob = emission[t - 1, tokens[j - 1] if changed > stayed else 0].exp().item()
        path.append(Point(j - 1, t - 1, prob))

        if changed > stayed:
            j -= 1
            if j == 0:
                break
    else:
        return None
    return path[::-1]


def _merge_repeats(path: List[Point], transcript: str) -> List[Segment]:
    """Merges repeated tokens in the alignment path.

    Args:
        path (List[Point]): The alignment path.
        transcript (str): The transcript text.

    Returns:
        List[Segment]: The merged segments.
    """
    i1, i2 = 0, 0
    segments = []
    while i1 < len(path):
        while i2 < len(path) and path[i1].token_index == path[i2].token_index:
            i2 += 1
        score = sum(path[k].score for k in range(i1, i2)) / (i2 - i1)
        segments.append(
            Segment(
                transcript[path[i1].token_index],
                path[i1].time_index,
                path[i2 - 1].time_index + 1,
                score,
            )
        )
        i1 = i2
    return segments


def _interpolate_nans(x: pd.Series, method: str = "nearest") -> pd.Series:
    """Interpolates NaN values in a pandas Series.

    Args:
        x (pd.Series): The pandas Series.
        method (str): The interpolation method (default: "nearest").

    Returns:
        pd.Series: The Series with interpolated NaNs.
    """
    if x.notnull().sum() > 1:
        return x.interpolate(method=method).ffill().bfill()
    else:
        return x.ffill().bfill()


def _align_segments(
    transcript: List[SingleSegment],
    model: torch.nn.Module,
    model_dictionary: Dict[str, int],
    model_lang: Language,
    model_type: str,
    audio: Audio,
    device: DeviceType,
    max_duration: float,
    return_char_alignments: bool,
    interpolate_method: str,
) -> Tuple[List[SingleAlignedSegment], List[SingleWordSegment]]:
    """Align segments based on the predictions.

    Args:
        transcript (List[SingleSegment]): The list of transcription segments.
        model (torch.nn.Module): The alignment model.
        model_dictionary (Dict[str, int]): Dictionary for character indices.
        model_lang (str): Language of the model.
        model_type (str): The type of the model ('torchaudio' or 'huggingface').
        audio (Audio): The audio data.
        device (DeviceType): The device to run the model on.
        max_duration (float): Maximum duration of the audio.
        return_char_alignments (bool): Flag to return character alignments.
        interpolate_method (str): Method for interpolating NaNs.

    Returns:
        Tuple[List[SingleAlignedSegment], List[SingleWordSegment]]: The aligned segments and word segments.
    """
    aligned_segments: List[SingleAlignedSegment] = []
    word_segments: List[SingleWordSegment] = []

    for sdx, segment in enumerate(transcript):
        t1 = segment["start"]
        t2 = segment["end"]
        text = segment["text"]

        aligned_segment: SingleAlignedSegment = {"start": t1, "end": t2, "text": text, "words": [], "chars": None}

        if return_char_alignments:
            aligned_segment["chars"] = []

        if _can_align_segment(segment, model_dictionary, t1, max_duration):
            _align_single_segment(
                segment,
                model,
                model_dictionary,
                model_lang,
                model_type,
                audio,
                device,
                t1,
                t2,
                return_char_alignments,
                interpolate_method,
                aligned_segment,
                aligned_segments,
                word_segments,
            )
        else:
            print(f'Failed to align segment ("{segment["text"]}"), skipping...')
            aligned_segments.append(aligned_segment)

    return (aligned_segments, word_segments)


def _convert_to_scriptline(data: AlignedTranscriptionResult) -> List[ScriptLine]:
    """Convert a dictionary of segments and word segments to a list of ScriptLine objects.

    Args:
        data (AlignedTranscriptionResult): The input dictionary with segments and word segments.

    Returns:
        List[ScriptLine]: The list of ScriptLine objects.
    """
    segments = data["segments"]
    script_lines = []

    for segment in segments:
        words = segment["words"]
        word_chunks = [ScriptLine(text=word["word"]) for word in words]

        # Handle 'nan' end values by setting them to None
        start = segment["start"]
        end: Optional[float] = segment["end"]
        if end is not None and (isinstance(end, float) and math.isnan(end)):
            end = None

        script_line = ScriptLine(text=segment["text"], start=start, end=end, chunks=word_chunks)
        script_lines.append(script_line)

    return script_lines


def _align_transcription(
    transcript: List[SingleSegment],
    model: torch.nn.Module,
    align_model_metadata: Dict[str, Any],
    audio: Audio,
    device: DeviceType,
    interpolate_method: str = "nearest",
    return_char_alignments: bool = False,
    print_progress: bool = False,
    combined_progress: bool = False,
) -> AlignedTranscriptionResult:
    """Aligns phoneme recognition predictions to known transcription.

    Args:
        transcript (List[SingleSegment]): The list of transcription segments.
        model (torch.nn.Module): The alignment model.
        align_model_metadata (Dict[str, Any]): Metadata for the alignment model.
        audio (Audio): The audio data.
        device (DeviceType): The device to run the model on.
        interpolate_method (str): The method for interpolating NaNs (default: "nearest").
        return_char_alignments (bool): Whether to return character alignments (default: False).
        print_progress (bool): Whether to print progress (default: False).
        combined_progress (bool): Whether to combine progress (default: False).

    Returns:
        AlignedTranscriptionResult: The aligned transcription result.
    """
    max_duration = audio.waveform.shape[1] / audio.sampling_rate

    model_dictionary = align_model_metadata["dictionary"]
    model_lang = align_model_metadata["language"]
    model_type = align_model_metadata["type"]

    transcript = _preprocess_segments(
        transcript,
        align_model_metadata["dictionary"],
        align_model_metadata["language"],
        print_progress,
        combined_progress,
    )

    aligned_segments, word_segments = _align_segments(
        transcript=transcript,
        model=model,
        model_dictionary=model_dictionary,
        model_lang=model_lang,
        model_type=model_type,
        audio=audio,
        device=device,
        max_duration=max_duration,
        return_char_alignments=return_char_alignments,
        interpolate_method=interpolate_method,
    )
    return {"segments": aligned_segments, "word_segments": word_segments}


def align_transcriptions(
    audios_and_transcriptions_and_language: List[Tuple[Audio, ScriptLine, Language]],
) -> List[List[ScriptLine]]:
    """Aligns transcriptions with the given audio using a wav2vec2.0 model.

    Args:
        audios_and_transcriptions_and_language (List[tuple[Audio, ScriptLine, Language]):
            A list of tuples audios, corresponding transcriptions, and optionally a language.
            The default language is English.

    Returns:
        List[List[ScriptLine]]: A list of lists, where each inner list contains the aligned script lines for each audio.
    """
    aligned_script_lines = []
    loaded_processors_and_models = {}

    for item in audios_and_transcriptions_and_language:
        audio, transcription, language = (*item, None)[:3]

        # Set default language to English if not provided
        if language is None:
            language = Language(language_code="en")

        # Define the language code and load model
        device = _select_device_and_dtype()[0]  # DeviceType object
        model_variant = DEFAULT_ALIGN_MODELS_HF.get(language.language_code, DEFAULT_ALIGN_MODELS_HF["en"])

        if model_variant.path_or_uri not in loaded_processors_and_models:  # Load model
            processor = Wav2Vec2Processor.from_pretrained(model_variant.path_or_uri)
            model = Wav2Vec2ForCTC.from_pretrained(model_variant.path_or_uri).to(device.value)
            loaded_processors_and_models[model_variant.path_or_uri] = (processor, model)

        processor, model = loaded_processors_and_models[model_variant.path_or_uri]

        if audio.sampling_rate != SAMPLE_RATE:
            raise ValueError(f"{audio.sampling_rate} rate is not equal to the training rate of {SAMPLE_RATE}.")

        # Ensure start and end are not None
        start = transcription.start if transcription.start is not None else 0.0
        end = transcription.end if transcription.end is not None else audio.waveform.shape[1] / audio.sampling_rate

        # Ensure text is not None
        text = transcription.text if transcription.text is not None else ""

        # Align each segment of the transcription
        segments = [
            SingleSegment(
                start=start, end=end, text=text, clean_char=None, clean_cdx=None, clean_wdx=None, sentence_spans=None
            )
        ]

        with threadpool_limits(limits=1, user_api="blas"):
            alignment = _align_transcription(
                transcript=segments,
                model=model,
                align_model_metadata={
                    "dictionary": processor.tokenizer.get_vocab(),
                    "language": language,
                    "type": "huggingface" if isinstance(model_variant, HFModel) else "torchaudio",
                },
                audio=audio,
                device=device,
                return_char_alignments=True,
            )
            aligned_script_lines.append(_convert_to_scriptline(alignment))

    return aligned_script_lines


# Note: this code is derived from: https://github.com/m-bain/whisperX/tree/main

# Copyright (c) 2022, Max Bain
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Max Bain.
# 4. Neither the name of Max Bain nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



"""Data structures relevant for handling audio files and metadata.

The most basic unit is the Audio object which represents the necessary information of a loaded audio
file and its corresponding metadata. Other functionality and abstract data types are provided for
ease of maintaining the codebase and offering consistent public APIs.
"""

import os
import uuid
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
import torchaudio
from pydantic import BaseModel, Field, ValidationInfo, field_validator

from senselab.utils.constants import SENSELAB_NAMESPACE


class Audio(BaseModel):
    """Pydantic model for audio and its corresponding metadata.

    Pydantic model for audio that holds the necessary attributes, the actual decoded audio data
    and the sampling rate, to work with audio in python. Contains metadata information as needed
    and has a unique identifier for every audio.

    Attributes:
        waveform: The actual audio data read from an audio file, stored as a torch.Tensor
            of shape (num_channels, num_samples)
        sampling_rate: The sampling rate of the audio file
        orig_path_or_id: Optional str for the original path or an ID to track file over time
        metadata: Optional metadata dictionary of information associated with this Audio instance
            (e.g. participant demographics, audio settings, location information)
    """

    waveform: torch.Tensor
    sampling_rate: int
    orig_path_or_id: Optional[str] = None
    metadata: Dict = Field(default={})
    model_config = {"arbitrary_types_allowed": True}

    @field_validator("waveform", mode="before")
    def convert_to_tensor(
        cls, v: Union[List[float], List[List[float]], np.ndarray, torch.Tensor], _: ValidationInfo
    ) -> torch.Tensor:
        """Converts the audio data to torch.Tensor of shape (num_channels, num_samples)."""
        temporary_tensor = None
        if isinstance(v, list):
            temporary_tensor = torch.tensor(v)
        elif isinstance(v, np.ndarray):
            temporary_tensor = torch.tensor(v)
        elif isinstance(v, torch.Tensor):
            temporary_tensor = v
        else:
            raise ValueError("Unsupported data type")

        if len(temporary_tensor.shape) == 1:
            # make the audio data [channels=1, samples]
            temporary_tensor = temporary_tensor.unsqueeze(0)
        return temporary_tensor.to(torch.float32)

    @classmethod
    def from_filepath(cls, filepath: str, metadata: Dict = {}) -> "Audio":
        """Creates an Audio instance from an audio file.

        Args:
            filepath: Filepath of the audio file to read from
            metadata: Additional information associated with the audio file
        """
        array, sampling_rate = torchaudio.load(filepath)

        return cls(waveform=array, sampling_rate=sampling_rate, orig_path_or_id=filepath, metadata=metadata)

    def generate_path(self) -> str:
        """Generate a path like string for this Audio.

        Generates a path like string for the Audio by either utilizing the orig_path_or_id, checking
        if it is a path (has an extension), otherwise using the id if orig_path_or_id is not an ID
        and giving an extension and relative to the current directory.
        """
        if self.orig_path_or_id:
            if os.path.splitext(self.orig_path_or_id)[-1].lower():
                return self.orig_path_or_id
            else:
                return f"{self.orig_path_or_id}"
        else:
            return f"{self.id()}"

    def id(self) -> str:
        """Generate a unique identifier for the Audio.

        Generate a unique identifier for the Audio where equivalent waveforms and sampling
        rates generate the same IDs.

        Returns: String UUID of the Audio generated by an MD5 hash of the waveform and the sampling_rate
        """
        return str(uuid.uuid3(uuid.uuid3(SENSELAB_NAMESPACE, str(self.waveform)), str(self.sampling_rate)))

    def __eq__(self, other: object) -> bool:
        """Overloads the default BaseModel equality to correctly check equivalence, ignoring metadata."""
        if isinstance(other, Audio):
            return self.id() == other.id()
        return False


def batch_audios(audios: List[Audio]) -> Tuple[torch.Tensor, Union[int, List[int]], List[Dict]]:
    """Batches the Audios together into a single Tensor, keeping individual Audio information separate.

    Batch all of the Audios into a single Tensor of shape (len(audios), num_channels, num_samples).
    Keeps the Audio information related to each sampling rate and metadata separate for each Audio to
    allow for unbatching after running relevant functionality.

    Args:
        audios: List of audios to batch together. NOTE: Should all have the same number of channels
            and is generally advised to have the same sampling rates if running functionality
            that relies on the sampling rate.

    Returns:
        Returns a tuple of a Tensor that will have the shape (len(audios), num_channels, num_samples),
        the sampling rate (an integer if all have the same sampling rate), and a list of each individual
        audio's metadata information.

    Raises:
        RuntimeError: if all of the Audios do not have the same number of channels
    """
    sampling_rates = []
    batched_audio = []
    metadatas = []
    for audio in audios:
        sampling_rates.append(audio.sampling_rate)
        batched_audio.append(audio.waveform)
        metadatas.append(audio.metadata)

    return_sampling_rates: List[int] | int = int(sampling_rates[0]) if len(set(sampling_rates)) == 1 else sampling_rates

    return torch.stack(batched_audio), return_sampling_rates, metadatas


def unbatch_audios(batched_audio: torch.Tensor, sampling_rates: int | List[int], metadatas: List[Dict]) -> List[Audio]:
    """Unbatches Audios into a List of Audio objects.

    Uses the batched Audios, their respective sampling rates, and their corresponding metadatas to create
    a list of Audios.

    Args:
        batched_audio: torch.Tensor of shape (batch_size, num_channels, num_samples) to unstack
        sampling_rates: The sampling rate of each batched audio if they differ or a single sampling rate for all of them
        metadatas: The respective metadata for each of the batched audios

    Returns:
        List of Audio objects representing each of the Audios that were previously batched together

    Raises:
        ValueError if the batched_audio is not in the correct shape or if the number of batched_audios does not
            match the amount of metadata and sampling rates (if they were provided as a List) that were provided.
    """
    if len(batched_audio.shape) != 3:
        raise ValueError("Expected batched audios to be of shape (batch_size, num_channels, samples)")
    elif batched_audio.shape[0] != len(metadatas) or (
        isinstance(sampling_rates, List) and batched_audio.shape[0] != len(sampling_rates)
    ):
        raise ValueError(
            "Expected sizes of batched_audio, sampling_rates (if provided as a litst) \
                         and metadata to be equal"
        )

    audios = []
    for i in range(len(metadatas)):
        sampling_rate = sampling_rates[i] if isinstance(sampling_rates, List) else sampling_rates
        metadata = metadatas[i]
        audio = batched_audio[i]
        audios.append(Audio(waveform=audio, sampling_rate=sampling_rate, metadata=metadata))
    return audios



"""This module provides the implementation of dimensionality reduction.

Dimensionality reduction is a crucial preprocessing step in many machine learning and data analysis pipelines.
It aims to project a high-dimensional dataset into a lower-dimensional subspace while preserving as much of the
significant structure and information as possible. This process is particularly useful for visualization,
noise reduction, and improving computational efficiency in downstream tasks.

This module specifically implements three popular dimensionality reduction techniques: Principal Component Analysis
(PCA), t-Distributed Stochastic Neighbor Embedding (t-SNE), and Uniform Manifold Approximation and Projection (UMAP).

### Task Overview
- Input: High-dimensional data (typically represented as a matrix or tensor), i.e. A tensor of shape (100, 10)
- Output: Lower-dimensional representation of the input data, i.e. A tensor of shape (100, 2)
- Goal: Preserve important structures or relationships in the data while reducing its dimensionality

### Evaluation Metrics
The effectiveness of dimensionality reduction techniques can be evaluated using several metrics, including but not
limited to:
- **Reconstruction Error**: Measures the difference between the original data and its reduced representation when
    projected back to the original space. Commonly used for techniques like PCA.
- **Trustworthiness**: Assesses how much the model preserves neighbor relationships between the high-dimensional
    space and the low-dimensional space.
- **Continuity**: Evaluates how well the local structure of the data is preserved in the reduced space.

### Supported models
- **Principal Component Analysis (PCA)**: PCA is a widely recognized technique in the field of data science
    and machine learning. It excels in linear dimensionality reduction, making it an invaluable tool for
    simplifying complex datasets. PCA's primary strength lies in its ability to capture the most significant
    variance within the data. By identifying the principal components that account for the majority of the variation,
    PCA allows researchers and analysts to reduce the dimensionality of their data while retaining its most important
    features. This makes PCA particularly useful for tasks like data visualization and feature extraction.

- **t-Distributed Stochastic Neighbor Embedding (t-SNE)**: t-SNE is primarily utilized as a visualization technique,
    setting it apart from more general dimensionality reduction methods. It is able to preserve local structures
    within data when projecting it into lower-dimensional spaces, typically two or three dimensions. This property
    makes t-SNE exceptionally good at creating visually interpretable embeddings that can reveal clusters and patterns
    in complex datasets. However, t-SNE is not suitable for tasks that require the preservation of global structures.

- **Uniform Manifold Approximation and Projection (UMAP)**: UMAP is a relatively newer technique that has gained
    significant attention due to its versatility and performance. UMAP is able to balance the preservation of both
    global and local structures within data. Like t-SNE, UMAP can be used visualizing high-dimensional data in two
    or three dimensions. However, UMAP is also effective as a general-purpose non-linear dimensionality reduction
    technique, making it suitable for a wide range of applications where preserving the overall structure of the
    data in lower dimensions is crucial.

#### Summary:
.. include:: ./dimensionality_reduction_table.md

References:
- PCA: https://www.nature.com/articles/s43586-022-00184-w
- t-SNE: https://jmlr.org/papers/v9/vandermaaten08a.html
- UAMP: https://arxiv.org/abs/1802.03426

### Notes:
This module leverages the implementations from scikit-learn for PCA and t-SNE, and the umap-learn library for UMAP.
The reduced data is returned as a PyTorch tensor, allowing seamless integration with PyTorch-based workflows.

Learn more:
- PCA: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
- t-SNE: https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
- UMAP: https://umap-learn.readthedocs.io/
"""

from typing import Literal

import pydra
import torch
import umap
from sklearn.decomposition import PCA as sklearn_PCA
from sklearn.manifold import TSNE as sklearn_TSNE


def compute_dimensionality_reduction(
    data: torch.Tensor, model: Literal["pca", "tsne", "umap"] = "pca", n_components: int = 2, **kwargs: object
) -> torch.Tensor:
    """Reduce the dimensionality of the given data using the specified model.

    This function provides three methods: PCA, t-SNE, or UMAP. The reduced data is returned as a PyTorch tensor.

    Args:
        data (torch.Tensor): Input data tensor of shape (n_samples, n_features).
        model ({"pca", "tsne", "umap"}, optional): The dimensionality reduction model to use.
            Choices are:
                "pca" for Principal Component Analysis (default),
                "tsne" for t-Distributed Stochastic Neighbor Embedding, or
                "umap" for Uniform Manifold Approximation and Projection
        n_components (int, optional): Number of dimensions in the output.
            Must be less than or equal to the number of features in the input data.
            Defaults to 2.
        **kwargs: Additional keyword arguments to pass to the chosen model.
            (See below for links to each model's documentation for available parameters.)

    Returns:
        torch.Tensor: The reduced data tensor of shape (n_samples, n_components).

    Raises:
        ValueError: If an invalid model choice is provided or if n_components is
            greater than the number of features in the input data or less than 0.

    Examples:
        >>> data = torch.randn(100, 10)  # 100 samples, 10 features
        >>> data.shape
        torch.Size([100, 10])

        >>> reduced_data = compute_dimensionality_reduction(data)
        >>> print(reduced_data.shape)
        torch.Size([100, 2])

        >>> reduced_data = compute_dimensionality_reduction(data, model="pca", n_components=2, svd_solver="full")
        >>> print(reduced_data.shape)
        torch.Size([100, 2])

        >>> reduced_data = compute_dimensionality_reduction(data, model="tsne", n_components=3, perplexity=30)
        >>> print(reduced_data.shape)
        torch.Size([100, 3])

        >>> reduced_data = compute_dimensionality_reduction(data, model="umap", n_components=4,
            force_approximation_algorithm=False, init='spectral', learning_rate=1.0, n_neighbors=5)
        >>> print(reduced_data.shape)
        torch.Size([100, 4])

    Notes:
        This function uses implementations from scikit-learn for PCA and t-SNE, and the umap-learn
        library for UMAP. For detailed information about each method and its parameters, please
        refer to the following documentation:

        - PCA: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
        - t-SNE: https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
        - UMAP: https://umap-learn.readthedocs.io/en/latest/api.html
    """
    if n_components > data.shape[1]:
        raise ValueError("n_components must be less than or equal to the number of features in the input data")

    if n_components < 1:
        raise ValueError("n_components must be greater than 0")

    if data.numel() == 0:
        raise ValueError("Input data must not be empty")

    if model == "pca":
        # Perform PCA
        reducer = sklearn_PCA(n_components=n_components, **kwargs)
        reduced_data = reducer.fit_transform(data)

    elif model == "tsne":
        # Perform t-SNE
        reducer = sklearn_TSNE(n_components=n_components, **kwargs)
        reduced_data = reducer.fit_transform(data)

    elif model == "umap":
        # Perform UMAP
        reducer = umap.UMAP(n_components=n_components, **kwargs)
        reduced_data = reducer.fit_transform(data)

    else:
        raise ValueError(f"Invalid model choice: {model}")

    return torch.from_numpy(reduced_data).float()


compute_dimensionality_reduction_pt = pydra.mark.task(compute_dimensionality_reduction)



"""Utility functions for utilizing different devices in Senselab."""

from enum import Enum
from typing import Optional

import torch


class DeviceType(Enum):
    """Device types for PyTorch operations."""

    CPU: str = "cpu"
    CUDA: str = "cuda"
    MPS: str = "mps"


DTYPE_MAP = {DeviceType.CPU: torch.float32, DeviceType.CUDA: torch.float16, DeviceType.MPS: torch.float32}


def _select_device_and_dtype(
    user_preference: Optional[DeviceType] = None,
    compatible_devices: list[DeviceType] = [
        DeviceType.CPU,
        DeviceType.CUDA,
        DeviceType.MPS,
    ],
) -> tuple[DeviceType, torch.dtype]:
    """Determines the device and data type for PyTorch operations.

    Allows users to give preferences for DeviceType, but determines based
    on compatible and available devices. Chooses the fastest option if no
    user preference is given.

    Args:
        user_preference: Optional DeviceType that the user wants to use
        compatible_devices: DeviceTypes that work with the functionality of the method calling this
    Returns:
        Tuple of (DeviceType, torch.dtype) where the device is both available and compatible and the
            dtype is the best performing dtype for that DeviceType
    Raises:
        ValueError: if the user specifies a preference that is not available or compatible and a safety
            call if no devices are available or compatible (we believe this to be impossible to trigger).
    """
    if user_preference:
        if not isinstance(user_preference, DeviceType):
            raise ValueError(f"user_preference should be of type DeviceType, not {type(user_preference)}")
    available_devices = [DeviceType.CPU]
    if torch.cuda.is_available():
        available_devices.append(DeviceType.CUDA)

    if torch.backends.mps.is_available():
        available_devices.append(DeviceType.MPS)

    # Check compatible and available
    useable_devices = []
    for device in available_devices:
        if device in compatible_devices:
            useable_devices.append(device)

    # User preference or fastest option

    if user_preference:
        # user_preference = DeviceType(user_preference) if isinstance(user_preference,str) else user_preference
        if user_preference not in useable_devices:
            raise ValueError(
                "The requested DeviceType is either not available or\
                             compatible with this functionality."
            )
        else:
            return user_preference, DTYPE_MAP[user_preference]
    else:
        if DeviceType.CUDA in useable_devices:
            return DeviceType.CUDA, DTYPE_MAP[DeviceType.CUDA]
        elif DeviceType.MPS in useable_devices:
            return DeviceType.MPS, DTYPE_MAP[DeviceType.MPS]
        elif DeviceType.CPU in useable_devices:
            return DeviceType.CPU, DTYPE_MAP[DeviceType.CPU]
        else:
            raise ValueError("Something went wrong and no devices were available or compatible.")



"""This module implements pooling methods for torch tensors."""

from typing import Literal

import pydra
import torch


def pooling(
    data: torch.Tensor,
    pool_type: Literal["mean", "min", "max"] = "mean",
    dimension: int = 0,
    keep_dimension: bool = False,
) -> torch.Tensor:
    """Apply pooling to the input tensor along the specified dimension.

    This function provides three pooling methods: mean, min, or max. The pooled tensor is returned as a PyTorch tensor.

    Args:
        data (torch.Tensor): Input tensor of shape (n_samples, n_features, n_channels).
        pool_type (str, optional): The pooling method to use.
            Choices are:
                * "mean" for mean pooling
                * "min" for min pooling
                * "max" for max pooling
            Defaults to "mean".
        dimension (int, optional): The dimension along which to apply the pooling.
            Defaults to 0.
        keep_dimension (bool, optional): Whether to keep the original tensor dimension after pooling.

    Returns:
        torch.Tensor: The pooled tensor of shape (n_samples, n_features, n_channels).

    Raises:
        ValueError: If an invalid pooling method choice is provided.

    Examples:
        >>> data = torch.randn(10, 20, 30)  # 10 samples, 20 features, 30 channels
        >>> data.shape
        torch.Size([10, 20, 30])

        >>> pooled_data = pooling(data, pool_type="mean", dimension=0, keep_dimension=True)
        >>> print(pooled_data.shape)
        torch.Size([1, 20, 30])

        >>> pooled_data = pooling(data, pool_type="max", dimension=1, keep_dimension=False)
        >>> print(pooled_data.shape)
        torch.Size([10, 30])

        >>> pooled_data = pooling(data, pool_type="min", dimension=2)
        >>> print(pooled_data.shape)
        torch.Size([10, 20])
    """
    # Ensure the dimension is valid
    assert dimension < len(data.shape), "Invalid dimension provided."

    if data.numel() == 0:
        raise ValueError("Input data must not be empty")

    # Applying the appropriate pooling based on the pool type
    if pool_type == "max":
        return torch.max(data, dim=dimension, keepdim=keep_dimension).values
    elif pool_type == "mean":
        return torch.mean(data, dim=dimension, keepdim=keep_dimension)
    elif pool_type == "min":
        return torch.min(data, dim=dimension, keepdim=keep_dimension).values
    else:
        raise ValueError("Unsupported pooling type. Choose 'max', 'avg', or 'min'.")


pooling_pt = pydra.mark.task(pooling)



"""This module implements some IOTask utilities.

This modules is deprecated and will be removed soon.
"""

import os
from typing import Any, Dict, List, Union

import pydra
from datasets import Audio, Dataset, Image, load_dataset

from senselab.utils.data_structures.file import File, from_strings_to_files
from senselab.utils.data_structures.model import check_hf_repo_exists


def read_files_from_disk(files: Union[str, List[str]]) -> Dict[str, Any]:
    """Read files from disk and create a Hugging Face `Dataset` object."""
    if isinstance(files, str):
        files = [files]
    formatted_files = from_strings_to_files(files)

    def _from_files_to_dataset(files: List[File]) -> Dataset:
        """Reading files from disk and create a HuggingFace `Dataset` object."""
        # Checking if all files are of the same type
        if not all(file.type == files[0].type for file in files):
            raise ValueError("All files must be of the same type.")

        # Loading file paths
        file_data = [str(file.filepath) for file in files]

        # Creating the Dataset object
        dict_obj = {files[0].type: file_data}
        # Using the type of the first file as the key

        return _from_dict_to_hf_dataset(dict_obj)

    dataset = _from_files_to_dataset(formatted_files)
    return _from_hf_dataset_to_dict(dataset)


def read_dataset_from_disk(input_path: str, split: str, streaming: bool = False) -> Dict[str, Any]:
    """Loads a Hugging Face `Dataset` object from disk.

    It determines the format based on the file extension or directory.
    """
    # Determine the input format.
    if os.path.isdir(input_path):
        input_format = "arrow"
    else:
        input_format = os.path.splitext(input_path)[1].strip(".")

    # Load the dataset
    try:
        dataset = load_dataset(
            input_format,
            split=split,
            data_files=input_path,
            streaming=streaming,
        )
        return _from_hf_dataset_to_dict(dataset)
    except Exception as e:
        # Generic error handling, e.g., network issues, data loading issues
        raise RuntimeError(f"An error occurred while loading the dataset: {str(e)}")


def read_dataset_from_hub(
    remote_repository: str,
    revision: str = "main",
    split: str = "all",
) -> Dict[str, Any]:
    """Loads a Hugging Face `Dataset` object from the Hugging Face Hub.

    It includes support for private repositories.
    """
    if not check_hf_repo_exists(remote_repository, "main", "dataset"):
        raise RuntimeError(
            f"The repository {remote_repository} - {revision} - {split}" " does not exist or could not be accessed."
        )

    # Load the dataset
    try:
        dataset = load_dataset(path=remote_repository, revision=revision, split=split)
    except Exception as e:
        # Generic error handling, e.g., network issues, data loading issues
        raise RuntimeError(f"An error occurred while loading the dataset: {str(e)}")

    return _from_hf_dataset_to_dict(dataset)


def push_dataset_to_hub(
    dataset: Dict[str, Any], remote_repository: str, revision: str = "main", split: str = "all"
) -> None:
    """Uploads a Hugging Face `Dataset` object to the Hugging Face Hub."""
    hf_dataset = _from_dict_to_hf_dataset(dataset)

    try:
        # Upload the dataset to the Hugging Face Hub
        hf_dataset.push_to_hub(repo_id=remote_repository, revision=revision, split=split)
    except Exception as e:
        raise RuntimeError(f"Failed to push dataset to the hub: {str(e)}")
    return


def save_dataset_to_disk(
    dataset: Dict[str, Any],
    output_directory: str,
    output_name: str,
    output_format: str = "parquet",
) -> None:
    """Saves a Hugging Face `Dataset` object to disk."""
    # TODO: optimize saving process (playing with batch size and num of cpus)
    hf_dataset = _from_dict_to_hf_dataset(dataset)

    # Prepare the output path differently if the format is Arrow
    if output_format == "arrow":
        output_path = os.path.join(output_directory, output_name)
        # No extension for Arrow, it's a directory
    else:
        output_path = os.path.join(output_directory, f"{output_name}.{output_format}")

    # Create the output directory, ignore error if it already exists
    os.makedirs(output_directory, exist_ok=True)

    if output_format == "parquet":

        def _save_hf_dataset_as_parquet(dataset: Dataset, output_path: str) -> None:
            """Saves a Hugging Face `Dataset` object to parquet format."""
            dataset.to_parquet(output_path)

        _save_hf_dataset_as_parquet(hf_dataset, output_path)
    elif output_format == "json":

        def _save_hf_dataset_as_json(dataset: Dataset, output_path: str) -> None:
            """Saves a Hugging Face `Dataset` object to json format."""
            dataset.to_json(output_path)

        _save_hf_dataset_as_json(hf_dataset, output_path)
    elif output_format == "csv":

        def _save_hf_dataset_as_csv(dataset: Dataset, output_path: str) -> None:
            """Saves a Hugging Face `Dataset` object to csv format."""
            dataset.to_csv(output_path)

        _save_hf_dataset_as_csv(hf_dataset, output_path)
    elif output_format == "sql":

        def _save_hf_dataset_as_sql(dataset: Dataset, output_path: str) -> None:
            """Saves a Hugging Face `Dataset` object to sql format."""
            dataset.to_sql(output_path)

        _save_hf_dataset_as_sql(hf_dataset, output_path)
    elif output_format == "arrow":

        def _save_hf_dataset_as_arrow(dataset: Dataset, output_path: str) -> None:
            """Saves a Hugging Face `Dataset` object in Apache Arrow format."""
            dataset.save_to_disk(output_path)

        _save_hf_dataset_as_arrow(hf_dataset, output_path)
    else:
        raise ValueError(f"Unsupported output format: {output_format}")
    return


def _from_hf_dataset_to_dict(dataset: Dataset) -> Dict[str, Any]:
    """Converts a Hugging Face `Dataset` object to a dictionary."""
    return dataset.to_dict()


def _from_dict_to_hf_dataset(
    data: Dict[str, Any],
    image_columns: list[str] = ["image"],
    audio_columns: list[str] = ["audio"],
) -> Dataset:
    """Converts a dictionary to a Hugging Face `Dataset` object.

    It casts image and audio columns.
    """
    dataset = Dataset.from_dict(data)

    # Cast image columns
    for column in image_columns:
        if column in dataset.column_names:
            dataset = dataset.cast_column(column, Image())

    # Cast audio columns
    for column in audio_columns:
        if column in dataset.column_names:
            dataset = dataset.cast_column(column, Audio(mono=False))

    return dataset


read_files_from_disk_pt = pydra.mark.task(read_files_from_disk)
read_dataset_from_disk_pt = pydra.mark.task(read_dataset_from_disk)
read_dataset_from_hub_pt = pydra.mark.task(read_dataset_from_hub)
push_dataset_to_hub_pt = pydra.mark.task(push_dataset_to_hub)
save_dataset_to_disk_pt = pydra.mark.task(save_dataset_to_disk)



"""This module contains functions that extract features from audio files using the PRAAT library."""

from typing import Dict, List

import parselmouth
import pydra
from parselmouth.praat import call

from senselab.audio.data_structures.audio import Audio


def get_audios_durations(audios: List[Audio]) -> List[Dict[str, float]]:
    """Returns the duration of the Audio objects."""

    def get_audio_duration(audio: Audio) -> Dict[str, float]:
        def _get_duration(sound: parselmouth.Sound) -> float:
            return call(sound, "Get total duration")

        waveform = audio.waveform
        sampling_rate = audio.sampling_rate
        sound = parselmouth.Sound(waveform, sampling_frequency=sampling_rate)
        duration = _get_duration(sound)
        return {"duration": duration}

    durations = []
    for audio in audios:
        durations.append(get_audio_duration(audio))
    return durations


def get_audios_f0_descriptors(
    audios: List[Audio],
    f0min: float,
    f0max: float,
    unit: str = "Hertz",
) -> List[Dict[str, float]]:
    """Returns the fundamental frequency descriptors of the audios."""

    def get_audio_f0(audio: Audio, f0min: float, f0max: float, unit: str) -> Dict[str, float]:
        def _to_pitch(sound: parselmouth.Sound, f0min: float, f0max: float) -> parselmouth.Pitch:
            return call(sound, "To Pitch", 0.0, f0min, f0max)

        def _get_mean_f0(pitch: parselmouth.Pitch, unit: str) -> float:
            return call(pitch, "Get mean", 0, 0, unit)

        def _get_std_dev_f0(pitch: parselmouth.Pitch, unit: str) -> float:
            return call(pitch, "Get standard deviation", 0, 0, unit)

        waveform = audio.waveform
        sampling_rate = audio.sampling_rate
        sound = parselmouth.Sound(waveform, sampling_frequency=sampling_rate)

        pitch = _to_pitch(sound, f0min, f0max)
        return {
            f"f0_mean_{unit}": _get_mean_f0(pitch, unit),
            f"f0_std_dev_{unit}": _get_std_dev_f0(pitch, unit),
        }

    return [get_audio_f0(audio, f0min, f0max, unit) for audio in audios]


def get_audios_harmonicity_descriptors(audios: List[Audio], f0min: float) -> List[Dict[str, float]]:
    """Returns the harmonicity descriptors of the audios."""

    def get_audio_harmonicity(audio: Audio, f0min: float) -> Dict[str, float]:
        def _to_harmonicity(sound: parselmouth.Sound, f0min: float) -> parselmouth.Harmonicity:
            return call(sound, "To Harmonicity (cc)", 0.01, f0min, 0.1, 1.0)

        def _get_mean_hnr(harmonicity: parselmouth.Harmonicity) -> float:
            return call(harmonicity, "Get mean", 0, 0)

        def _get_std_dev_hnr(harmonicity: parselmouth.Harmonicity) -> float:
            return call(harmonicity, "Get standard deviation", 0, 0)

        waveform = audio.waveform
        sampling_rate = audio.sampling_rate
        sound = parselmouth.Sound(waveform, sampling_frequency=sampling_rate)

        harmonicity = _to_harmonicity(sound, f0min)
        return {
            "harmonicity_mean": _get_mean_hnr(harmonicity),
            "harmonicity_std_dev": _get_std_dev_hnr(harmonicity),
        }

    return [get_audio_harmonicity(audio, f0min) for audio in audios]


def get_audios_jitter_descriptors(
    audios: List[Audio],
    f0min: float,
    f0max: float,
) -> List[Dict[str, float]]:
    """Returns the jitter descriptors of the audios."""

    def get_audio_jitter(audio: Audio, f0min: float, f0max: float) -> Dict[str, float]:
        def _to_point_process(sound: parselmouth.Sound, f0min: float, f0max: float) -> parselmouth.Data:
            return call(sound, "To PointProcess (periodic, cc)", f0min, f0max)

        def _get_jitter(type: str, point_process: parselmouth.Data) -> float:
            return call(point_process, f"Get jitter ({type})", 0, 0, 0.0001, 0.02, 1.3)

        waveform = audio.waveform
        sampling_rate = audio.sampling_rate
        sound = parselmouth.Sound(waveform, sampling_frequency=sampling_rate)
        point_process = _to_point_process(sound, f0min, f0max)
        return {
            "local_jitter": _get_jitter("local", point_process),
            "localabsolute_jitter": _get_jitter("local, absolute", point_process),
            "rap_jitter": _get_jitter("rap", point_process),
            "ppq5_jitter": _get_jitter("ppq5", point_process),
            "ddp_jitter": _get_jitter("ddp", point_process),
        }

    return [get_audio_jitter(audio, f0min, f0max) for audio in audios]


def get_audios_shimmer_descriptors(
    audios: List[Audio],
    f0min: float,
    f0max: float,
) -> List[Dict[str, float]]:
    """Returns the shimmer descriptors of the audios."""

    def get_audio_shimmer(audio: Audio, f0min: float, f0max: float) -> Dict[str, float]:
        def _to_point_process(sound: parselmouth.Sound, f0min: float, f0max: float) -> parselmouth.Data:
            return call(sound, "To PointProcess (periodic, cc)", f0min, f0max)

        def _get_shimmer(type: str, sound: parselmouth.Sound, point_process: parselmouth.Data) -> float:
            return call([sound, point_process], f"Get shimmer ({type})", 0, 0, 0.0001, 0.02, 1.3, 1.6)

        waveform = audio.waveform
        sampling_rate = audio.sampling_rate
        sound = parselmouth.Sound(waveform, sampling_frequency=sampling_rate)
        point_process = _to_point_process(sound, f0min, f0max)
        return {
            "local_shimmer": _get_shimmer("local", sound, point_process),
            "localDB_shimmer": _get_shimmer("local_dB", sound, point_process),
            "apq3_shimmer": _get_shimmer("apq3", sound, point_process),
            "apq5_shimmer": _get_shimmer("apq5", sound, point_process),
            "apq11_shimmer": _get_shimmer("apq11", sound, point_process),
            "dda_shimmer": _get_shimmer("dda", sound, point_process),
        }

    return [get_audio_shimmer(audio, f0min, f0max) for audio in audios]


get_audios_durations_pt = pydra.mark.task(get_audios_durations)
get_audios_f0_descriptors_pt = pydra.mark.task(get_audios_f0_descriptors)
get_audios_harmonicity_descriptors_pt = pydra.mark.task(get_audios_harmonicity_descriptors)
get_audios_jitter_descriptors_pt = pydra.mark.task(get_audios_jitter_descriptors)
get_audios_shimmer_descriptors_pt = pydra.mark.task(get_audios_shimmer_descriptors)
